/*
 *  File Name: setup_utils.c
 * 
 *  Description: Contains setup utilities (described in Chapter 2), color
 *               utilities (described in Chapter 3) and other miscellaneous
 *               goodies
 *
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/PEX5/PEXlib.h>
#ifdef EXT_XMU
#include <X11/Xmu/StdCmapExt.h>
#endif

/********************************************************************/
/*        Setup Utilities                                           */
/********************************************************************/

#define BORDER_WIDTH    (1)

Display *open_pex_connection(ext_info) 
PEXExtensionInfo **ext_info;
{
    Display *dpy;
    char error_string[PEXErrorStringLength];

    /* Open a connection to the X server */
    if ( !(dpy = XOpenDisplay( "")) ) {
       fprintf( stderr, "Cannot open display %s \n", XDisplayName(""));
       return NULL;
    }

    /* Initialize the PEX extension and get the PEX extension information  */
    /* We will need to know the version of PEX later on.                   */
    if ( PEXInitialize(dpy, ext_info, PEXErrorStringLength, error_string) != 0 ) {
	fprintf( stderr, "Error Initializing PEX: %s\n", error_string);
        XCloseDisplay( dpy);
        return NULL;
    }

    return dpy;
}

Window create_window(dpy, title_string, bg_pixel, vis_info, cmap_info, width, 
    height)
Display *dpy;
char *title_string;
unsigned long bg_pixel;
XVisualInfo *vis_info;
XStandardColormap *cmap_info;
unsigned int *width;
unsigned int *height;
{
    Window win; 
    int	screen;  /* default screen */
    unsigned int screen_width, screen_height;
    XSizeHints win_hint; /* window manager hints */
    XSetWindowAttributes win_attr;
    int	status;

    /* Compute the window size so it is one quarter of the screen */
    screen = DefaultScreen( dpy);
    screen_width = DisplayWidth( dpy, screen);
    screen_height = DisplayHeight( dpy, screen);

    win_hint.x = win_hint.y = 0;
    if ( screen_width > screen_height )
        *width = *height = win_hint.width = win_hint.height = screen_height/2;
    else
        *width = *height = win_hint.width = win_hint.height = screen_width/2;
    win_hint.flags = PPosition | PSize;

    /* Create the window using the specified visual and colormap      */
    win_attr.background_pixel = bg_pixel;
    win_attr.colormap = cmap_info->colormap;
    win = XCreateWindow( dpy, RootWindow(dpy, screen),
        win_hint.x, win_hint.y, win_hint.width, win_hint.height,
        BORDER_WIDTH, vis_info->depth, InputOutput, vis_info->visual,
        CWColormap | CWBackPixel, &win_attr);
   
   /* Now set the window properties. These are hints for the window manager */
   XSetStandardProperties( dpy, win, title_string, title_string, None, NULL, 
        0, &win_hint);

   return win;
}

PEXRenderer create_renderer(dpy, win, approx_entry, rd_flags, rd_attr)
Display *dpy;
Window win;
PEXColorApproxEntry *approx_entry;
unsigned long rd_flags;
PEXRendererAttributes *rd_attr;
{
    unsigned long flags;

    /* Set the color approximation to the one specified */
    rd_attr->color_approx_table =
          PEXCreateLookupTable( dpy, win, PEXLUTColorApprox);
    PEXSetTableEntries( dpy, rd_attr->color_approx_table, 0/*default entry*/,
          1/*count*/, PEXLUTColorApprox, approx_entry);
    flags = rd_flags | PEXRAColorApproxTable;

    return( PEXCreateRenderer( dpy, win, flags, rd_attr));
}

/********************************************************************/
/*        Color Utilities                                           */
/********************************************************************/

#define MAX_TARGETS     (100)

/* Data structure for SERVER_OVERLAY_VISUALS property */
/* Used by HP, DEC, SGI, Kubota and possibly others   */
typedef struct {
   VisualID visualid;
   int transparent_type; /* None, TransparentPixel, TransparentMask */
   int value; /* Pixel value */
   int layer; /* Overlay planes will be in layer 1 */
} OverlayProperty;

void get_overlay_visuals(dpy, noverlay_visuals, overlay_visuals)
Display *dpy;
int *noverlay_visuals;
OverlayProperty **overlay_visuals;
{
   Atom property_atom;
   Atom actual_type;
   int actual_format;
   unsigned long item_count;
   unsigned long num_items, bytes_unread;
   int result;
   int	screen;

   screen = DefaultScreen( dpy);

   /* If the atom doesn't exist then the property isn't there */
   property_atom = XInternAtom(dpy, "SERVER_OVERLAY_VISUALS", True);
   if ( property_atom == None ) {
      *noverlay_visuals = 0;
      *overlay_visuals = (OverlayProperty *) NULL;
      return;
   }

   /* Fetch the property. We want a list of overlay visuals */
   bytes_unread = 0;
   item_count = sizeof( OverlayProperty)/4;
   do {
      item_count += (bytes_unread + 3)/4;
      result = XGetWindowProperty( dpy, RootWindow( dpy, screen),
	 property_atom, 0, item_count, False, property_atom,
	 &actual_type, &actual_format, &num_items, &bytes_unread,
	 (unsigned char **) overlay_visuals );

   } while ( bytes_unread > 0 );

   *noverlay_visuals = num_items/4;
}

int is_overlay_visual(noverlay_visuals, overlay_visuals, visualid)
int noverlay_visuals;
OverlayProperty *overlay_visuals;
VisualID visualid;
{
   int i, status;

   status = False;
   for ( i = 0; i < noverlay_visuals; i++ ) {
     if ( visualid == overlay_visuals[i].visualid ) {
         status = True;
         break;
     }
   }

   return status;
}

int find_visual_5_0(dpy, vis_priorities, overlay, vis_return)
Display *dpy; 
int *vis_priorities;
int overlay;
XVisualInfo *vis_return;
{
   XVisualInfo sample_vis, *vis_list, *best_vis;
   int i, this_priority, best_priority, this_size, best_size, vis_count;
   int noverlay_visuals;
   OverlayProperty *overlay_visuals;
   int status;

   /* On 5.0 servers we cannot determine which visuals are pex-capable so we */
   /* must retrieve the entire list.                                         */
   sample_vis.screen = DefaultScreen( dpy);
   vis_list = XGetVisualInfo( dpy, VisualScreenMask, &sample_vis, &vis_count);

   /* Get the overlay visuals for this display */
   get_overlay_visuals( dpy, &noverlay_visuals, &overlay_visuals);

   /* We want to find visuals with priority greater that zero */
   best_priority = best_size = 0;
   best_vis = NULL;

   for (i = 0; i < vis_count; i++) {

      /* Check whether or not overlay visual is desired */
      if ( is_overlay_visual( noverlay_visuals, overlay_visuals, 
              vis_list[i].visualid) != overlay )
         continue;

      /* If the priority of this visual class is less than the best one so */
      /* far then continue; we do not want this visual                     */ 
      this_priority = vis_priorities[ vis_list[i].class ];
      if ( this_priority < best_priority ) continue;

      /* Get the number of colors for this visual */
      switch ( vis_list[i].class ) {
          case StaticGray: 
          case GrayScale:
          case StaticColor:
          case PseudoColor:
	      this_size = vis_list[i].colormap_size;
	      break;
          case TrueColor:
          case DirectColor:
	      this_size = vis_list[i].red_mask + vis_list[i].green_mask + 
	           vis_list[i].blue_mask + 1;
	  break;
      }

      /* If this visual class has a higher priority than the best visual   */
      /* so far then use this visual. If the priorites are equal then only */
      /* use this visual if it has more colors				   */
      if ( this_priority > best_priority || 
          (this_priority == best_priority && this_size > best_size) ) {
	 best_priority = this_priority; 
	 best_size = this_size;
	 best_vis = &vis_list[i];
      }
   }

   status = False;
   if ( best_vis ) {
      *vis_return = *best_vis;
      status = True;
   }

   if ( vis_count > 0 ) XFree ( vis_list);
   if ( noverlay_visuals > 0 ) XFree( overlay_visuals);
   return status;
}

int find_visual(dpy, vis_priorities, overlay, vis_return)
Display *dpy; 
int *vis_priorities;
int overlay;
XVisualInfo *vis_return;
{
   unsigned long vis_count;
   int i, this_priority, best_priority, this_size, best_size;
   PEXRenderingTarget *targets, *best_target;
   Visual *visual;
   int noverlay_visuals;
   OverlayProperty *overlay_visuals;
   int status;

   /* On 5.1 (or later) servers, PEX will tell us which visuals it supports */
   PEXMatchRenderingTargets( dpy, RootWindow( dpy, DefaultScreen(dpy)), 0, 
      PEXWindowDrawable, NULL, MAX_TARGETS, &vis_count, &targets);

   /* Get the overlay visuals for this display */
   get_overlay_visuals( dpy, &noverlay_visuals, &overlay_visuals);

   /* We want to find visuals with priority greater that zero */
   best_priority = best_size = 0;
   best_target = NULL;

   for ( i = 0; i < vis_count; i++ ) {
      visual = targets[i].visual;

      /* Check whether or not overlay visual is desired */
      if ( is_overlay_visual( noverlay_visuals, overlay_visuals, 
              visual->visualid) != overlay )
         continue;

      /* If the priority of this visual class is less than the best one so */
      /* far then continue; we do not want this visual                     */ 
      this_priority = vis_priorities[ visual->class ];
      if ( this_priority < best_priority ) continue;

      /* Get the number of colors for this visual */
      switch ( visual->class ) {
          case StaticGray: 
          case GrayScale:
          case StaticColor:
          case PseudoColor:
	      this_size = visual->map_entries;
	      break;
          case TrueColor:
          case DirectColor:
	      this_size = visual->red_mask + visual->green_mask + 
	           visual->blue_mask + 1;
	  break;
      }

      /* If this visual class has a higher priority than the best visual   */
      /* so far then use this visual. If the priorites are equal then only */
      /* use this visual if it has more colors				   */
      if ( this_priority > best_priority || 
          (this_priority == best_priority && this_size > best_size) ) {
	 best_priority = this_priority; 
	 best_size = this_size;
	 best_target = &targets[i];
      }
   }

   status = False;
   if ( best_target ) {
      vis_return->visual = best_target->visual;
      vis_return->visualid = best_target->visual->visualid;
      vis_return->class = best_target->visual->class;
      vis_return->red_mask = best_target->visual->red_mask;
      vis_return->green_mask = best_target->visual->green_mask;
      vis_return->blue_mask = best_target->visual->blue_mask;
      vis_return->colormap_size = best_target->visual->map_entries;
      vis_return->bits_per_rgb = best_target->visual->bits_per_rgb;
      vis_return->screen = DefaultScreen( dpy);
      vis_return->depth = best_target->depth;
      status = True;
   }

   if ( vis_count > 0 ) XFree( targets);
   if ( noverlay_visuals > 0 ) XFree( overlay_visuals);
   return status;
}

int find_colormap(dpy, vis_info, cmap_return)
Display *dpy; 
XVisualInfo *vis_info;
XStandardColormap *cmap_return;
{
   Atom wm_property;
   XStandardColormap *std_cmaps;
   int num_cmaps, i;
#ifdef EXT_XMU
   XmuParam extended_control;
   XmuExtendedParams extended_params;
#endif

   /* RGB Best map has the most colors and, thus, gives better results for   */
   /* shading. However, we only choose it on true color type visuals because */
   /* pseudo color visuals tend to have smaller colormaps.                   */
   switch ( vis_info->class ) {
      case StaticGray: 
      case GrayScale:
	 wm_property = XA_RGB_GRAY_MAP;
         break;
      case StaticColor:
      case PseudoColor:
         wm_property = XA_RGB_DEFAULT_MAP;
         break;
      case TrueColor:
      case DirectColor:
	 wm_property = XA_RGB_BEST_MAP;
         break;
   }

   /* Get the standard colormap attributes; if the specified colormap doesn't  */
   /* exist yet then it is created. Note, that we use an Xmu routine that was  */
   /* extended for PEX. Setting the PEX option causes it to load a colormap    */
   /* that will work with the PEX extension. (There is a special escape in PEX */
   /* to determine which color approx values--that is, the options for packing */
   /* r,g,b values into pixels--are supported. The pex option ensures that a   */
   /* supported color approx results.)                                         */
#ifdef EXT_XMU
   extended_control.name = XMU_CONTROL_PEX_SUPPORTED;
   extended_control.value = True;
   extended_params.num_controls = 1;
   extended_params.controls = &extended_control;
   extended_params.num_hints = 0;
   extended_params.hints = 0;
   if ( XmuLookupStandardColormapExt( dpy, vis_info->screen, vis_info->visualid,
	   vis_info->depth, wm_property, False/*replace*/, True/*retain*/,
           &extended_params) ) {
#else
   if ( XmuLookupStandardColormap( dpy, vis_info->screen, vis_info->visualid,
	   vis_info->depth, wm_property, False/*replace*/, True/*retain*/) ) {
#endif
      if ( XGetRGBColormaps( dpy, RootWindow( dpy, vis_info->screen), 
	   &std_cmaps, &num_cmaps, wm_property) ) {
         for (i = 0; i < num_cmaps; i++) {
	    if ( std_cmaps[i].visualid == vis_info->visualid ) {
               *cmap_return = std_cmaps[i];
               XFree( std_cmaps);
	       return True;
	    }
	 }
      }
   }

   return False;
}

int get_pex_color_attr(dpy, vis_info, cmap_info, color_approx)
Display *dpy; 
XVisualInfo *vis_info;
XStandardColormap *cmap_info;
PEXColorApproxEntry *color_approx;
{
   color_approx->model =  PEXColorApproxRGB;
   color_approx->base_pixel = cmap_info->base_pixel;
   color_approx->dither = (vis_info->depth < 24) ? PEXOn : PEXOff;

   /* Set up the color approx to match the colormap attributes */
   switch ( vis_info->class ) {
      case StaticGray: 
      case GrayScale:
         color_approx->type =  PEXColorRange;
         color_approx->max1 = cmap_info->red_max;
         /* max2 and max3 not used by color range */
         color_approx->max2 = 0;
         color_approx->max3 = 0;
         color_approx->mult1 = cmap_info->red_mult;
         color_approx->mult2 = 0;
         color_approx->mult3 = 0;
         /* use NTSC weights */
         color_approx->weight1 = .30;
         color_approx->weight2 = .59;
         color_approx->weight3 = .11;
         break;
      case StaticColor:
      case PseudoColor:
      case TrueColor:
      case DirectColor:
         color_approx->type = PEXColorSpace;
         color_approx->max1 = cmap_info->red_max;
         color_approx->max2 = cmap_info->green_max;
         color_approx->max3 = cmap_info->blue_max;
         color_approx->mult1 = cmap_info->red_mult;
         color_approx->mult2 = cmap_info->green_mult;
         color_approx->mult3 = cmap_info->blue_mult;
         /* weights not used by color space */
         color_approx->weight1 = 0;
         color_approx->weight2 = 0;
         color_approx->weight3 = 0;
         break;
   }

   return True;
}

int setup_colors(dpy, vis_info, cmap_info, color_attr)
Display *dpy; 
XVisualInfo *vis_info;
XStandardColormap *cmap_info; 
PEXColorApproxEntry *color_attr;
{
   int vis_priorities[6];
   int overlay;
   PEXExtensionInfo *ext_info;

   /* Tune this to your application. The supported visual that matches the    */
   /* overlay flag (this flag indicates whether or not the selected visual    */
   /* should be an overlay) and has the class with the highest (non-negative) */
   /* priority will be selected. (If a visual class has priority less than    */
   /* zero then visuals with that class will not be selected). If several     */
   /* visuals satisfy this criteria then the one that supports the most colors*/
   /* is selected. Note that if you disable certain visual classes or set the */
   /* overlay flag to True then find_visual may fail                          */

   /* We give highest priority to true-color visuals since this is the MOST */
   /* PORTABLE way to use PEX. We also disable selection of overlays        */
   vis_priorities[TrueColor] =
      vis_priorities[DirectColor] = 3;
   vis_priorities[PseudoColor] = 2;
   vis_priorities[StaticColor] = 
      vis_priorities[StaticGray] = 
      vis_priorities[GrayScale] = 1;
   overlay = False;

   /* See if this is a 5.0 or a 5.1 server */
   ext_info = PEXGetExtensionInfo( dpy);

   if ( ext_info->minor_version == 0 ) {
      if ( !find_visual_5_0( dpy, vis_priorities, overlay, vis_info) ) { 
         fprintf( stderr, "Cannot find suitable visual\n");
         return False;
      }
   }
   else {
      if ( !find_visual( dpy, vis_priorities, overlay, vis_info) ) { 
         fprintf( stderr, "Cannot find suitable visual\n");
         return False;
      }
   }

   if ( !find_colormap( dpy, vis_info, cmap_info) ) {
      fprintf( stderr, "Cannot set up colormap for visual %d\n", 
         vis_info->visualid);
      return False;
   }
  
   if ( !get_pex_color_attr( dpy, vis_info, cmap_info, color_attr) ) {
      fprintf( stderr, "Cannot set get color approximation for visual %d\n", 
         vis_info->visualid);
      return False;
   }

   return True;
}

/********************************************************************/
/*        Miscellaneous Utilities                                   */
/********************************************************************/

int check_enum( dpy, win, enum_type_name, enum_type)
Display *dpy;
Window win;
int enum_type_name;
int enum_type;
{
   unsigned long *enum_counts;
   PEXEnumTypeDesc *enum_desc;
   int i;
   int supported;

   /* Check if desired pick device type and method are supported */
   supported = False;
   if ( PEXGetEnumTypeInfo( dpy, win, 1, &enum_type_name, PEXETIndex,
          &enum_counts, &enum_desc) ) {
      for (i = 0; i < *enum_counts; i++) {
         if ( enum_desc[i].index == enum_type) {
             supported = True;
             break;
         }
      }
      PEXFreeEnumInfo( 1, enum_counts, enum_desc);
   }

   return ( supported);
}

