/*
 *  File Name: view_utils.c
 *
 *  Description: Contains utilities to set up a view and do double
 *               buffering
 * 
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <math.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/extensions/multibuf.h>
#include <X11/PEX5/PEXlib.h>
#include "esdb.h"
#include "view_utils.h"

#define MAX_TARGETS 100

static int error_flag = 0;
static int print_err;

/*
 * Set up view of object with either perspective or orthographic projection
 */
void init_object_view(dpy, xmin, xmax, ymin, ymax, zmin, zmax, scale, proj_type, 
     view_dir, view_index, view_table)
Display *dpy;
float xmin;
float xmax;
float ymin;
float ymax;
float zmin;
float zmax;
float scale;
int proj_type; /* ORTHOGRAPHIC_PROJECTION or PERSPECTIVE_PROJECTION */
int view_dir;  /* FRONT_VIEW, LEFT_SIDE_VIEW, RIGHT_SIDE_VIEW, TOP_VIEW */
int view_index;
PEXLookupTable view_table;
{
    PEXViewEntry view_entry;
    PEXCoord from;
    PEXCoord to;
    PEXVector up;
    float xside, yside, zside, view_side, inv_scale, near_plane, far_plane;

    /*
     * compute distance from center of bounding box to outer limits
     */
    xside = (xmax - xmin) / 2.0;
    yside = (ymax - ymin) / 2.0;
    zside = (zmax - zmin) / 2.0;

    /*
     * view_side is largest side of model extent rectangle
     * add 25% to this distance so view is not cramped
     */
    view_side = (xside > yside) ? xside : yside;
    if (zside > view_side) view_side = zside;
    view_side += view_side / 4.0; 

    /*
     * look from origin to center of object
     */
    switch (view_dir) {
       case FRONT_VIEW:
         to.z   = zmin + zside;
         from.z = to.z + view_side * 2;
         from.x = to.x = xmin + xside;
         from.y = to.y = ymin + yside;
         up.x   = 0.0;
         up.y   = 1.0;
         up.z   = 0.0;
         break;
       case BACK_VIEW:
         to.z   = zmin + zside;
         from.z = to.z - view_side * 2;
         from.x = to.x = xmin + xside;
         from.y = to.y = ymin + yside;
         up.x   = 0.0;
         up.y   = 1.0;
         up.z   = 0.0;
         break;
       case RIGHT_SIDE_VIEW:
         to.x   = xmin + xside;
         from.x = to.x + view_side * 2;
         from.y = to.y = ymin + yside;
         from.z = to.z = zmin + zside;
         up.x   = 0.0;
         up.y   = 1.0;
         up.z   = 0.0;
         break;
       case LEFT_SIDE_VIEW:
         to.x   = xmin + xside;
         from.x = to.x - view_side * 2;
         from.y = to.y = ymin + yside;
         from.z = to.z = zmin + zside;
         up.x   = 0.0;
         up.y   = 1.0;
         up.z   = 0.0;
         break;
       case TOP_VIEW:
         to.y   = ymin + yside;
         from.y = to.y + view_side * 2;
         from.x = to.x = xmin + xside;
         from.z = to.z = zmin + zside;
         up.x   = 0.0;
         up.y   = 0.0;
         up.z   = -1.0;
         break;
       case BOTTOM_VIEW:
         to.y   = ymin + yside;
         from.y = to.y - view_side * 2;
         from.x = to.x = xmin + xside;
         from.z = to.z = zmin + zside;
         up.x   = 0.0;
         up.y   = 0.0;
         up.z   = -1.0;
         break;
      case ISO_VIEW:
         to.x   = xmin + xside;
         to.y   = ymin + yside;
         to.z   = zmin + zside;
         from.x = to.x + view_side * 2;
         from.y = to.y + view_side * 2;
         from.z = to.z + view_side * 2;
         up.x   = 0.0;
         up.y   = 1.0;
         up.z   = 0.0;
         break;
    }

    PEXLookAtViewMatrix( &from, &to, &up, view_entry.orientation);

    /* 
     * Set near and far to be double the model extents. This ensures that
     * nothing will be clipped, even for rotated views such as ISO       
     */
    near_plane = view_side*2;
    far_plane = -view_side*2;

    /*
     * Set aspect ratio to 1 so original geometric proportions retained
     * For perspective set fov = 90 degrees. Note that
     *   half_width of window on viewing plane is view_side*inv_scale
     *   distance to front of object is half_width/tan(45) 
     *   distance to center of object is half_width/tan(45) + view_side 
     */
    inv_scale =  (scale == 0) ? (1.0) : (1 / scale);
    if (proj_type == ORTHOGRAPHIC_PROJECTION)
        PEXOrthoProjMatrix( view_side*2*inv_scale, 1.0, near_plane,
            far_plane, view_entry.mapping);
    else
        PEXPerspProjMatrix( M_PI/2.0, view_side*inv_scale + view_side, 
            1.0, near_plane, far_plane, view_entry.mapping);

    view_entry.clip_flags = 0;
    view_entry.clip_limits.min.x = view_entry.clip_limits.min.y = 
        view_entry.clip_limits.min.z = 0.0;
    view_entry.clip_limits.max.x = view_entry.clip_limits.max.y =
         view_entry.clip_limits.max.z = 1.0;

    PEXSetTableEntries( dpy, view_table, view_index, 1/*count*/,
        PEXLUTView, &view_entry);
}

int my_err_handler(dpy, myerr)
Display *dpy;
XErrorEvent *myerr;
{
    char msg[80];
    error_flag = myerr->error_code;
    XGetErrorText( dpy, myerr->error_code, msg, 80);
    if (print_err)
      fprintf( stderr, "Error code %s\n", msg);

    return;
}

int set_clear(dpy, rd, ext_info, background)
Display *dpy;
PEXRenderer rd;
PEXExtensionInfo *ext_info;
PEXColorRGB  *background;
{
   PEXRendererAttributes rd_attr;

   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb = *background;
      PEXChangeRenderer( dpy, rd, PEXRAClearImage | PEXRABackgroundColor, 
	&rd_attr);
      return True;
   } else
      return False;
}

int create_MBX_buffer(dpy, win, rd, ext_info, mbuffers)
Display *dpy;
Window  win;
PEXRenderer rd;
PEXExtensionInfo *ext_info;
Multibuffer *mbuffers;
{
   int major_num, event_num, error_num, n;

   if ( !XQueryExtension( dpy, MULTIBUFFER_PROTOCOL_NAME, &major_num, 
         &event_num, &error_num) ) {
      return( False);
   }

   /* Check if MBX drawable supported (this call was introduced in 5.1) */
   if ( ext_info->minor_version > 0 ) {
      XWindowAttributes win_attr;
      PEXRenderingTarget *targets;
      unsigned long count;

      XGetWindowAttributes( dpy, win, &win_attr);
      PEXMatchRenderingTargets( dpy, win, win_attr.depth, PEXBufferDrawable, 
	win_attr.visual, MAX_TARGETS, &count, &targets);
      XFree( targets);
      if (count <= 0) return False;
   }

   n = XmbufCreateBuffers( dpy, win, 2, MultibufferUpdateActionBackground,
           MultibufferUpdateHintFrequent, mbuffers);
   if (n < 2) {
      XmbufDestroyBuffers( dpy, win);
      return False;
   }

   /* On 5.0 servers check for support by trying to use it */
   if ( ext_info->minor_version == 0 ) {
      PEXBeginRendering( dpy, mbuffers[1], rd);
      XSync( dpy, 0);
      if (error_flag != 0) {
          error_flag = 0;
          PEXEndRendering( dpy, rd, False);
          XmbufDestroyBuffers( dpy, win);
          return False;
      }
      PEXEndRendering( dpy, rd, False);
   }

   return True;
}

int create_ES_buffer(dpy, win, rd, ext_info, do_clear)
Display *dpy;
Window  win;
PEXRenderer rd;
PEXExtensionInfo *ext_info;
int *do_clear;
{
   int enum_type_name;
   unsigned long *enum_counts;
   PEXEnumTypeDesc *enum_desc;
   int num_enums;
   int dblBuf, swapBuf, swapContent;
   esEscapeDblBuffer db_data;
   esEscapeSwapBufferContent swap_content_data;
   esEscapeSwapBufferContentReply *swap_content_reply;
   unsigned long reply_length;

   /* Don't use E&S double buffering on 5.0 servers */
   if ( ext_info->minor_version < 1 ) 
      return False;

   /* Only use E&S double buffering if all 3 escapes are supported */
   dblBuf = swapBuf = swapContent = False;
   enum_type_name = PEXETEscape;
   if ( PEXGetEnumTypeInfo( dpy, win, 1, &enum_type_name, PEXETIndex,
          &enum_counts, &enum_desc )) {
      num_enums = enum_counts[0];
      while (--num_enums >= 0) {
         switch(enum_desc[num_enums].index & 0xffff) {
           case PEXEncodeEnum(ES_VENDOR_ID, ESEscapeDoubleBufferID):
              dblBuf = True;
	      break;
            case PEXEncodeEnum(ES_VENDOR_ID, ESEscapeSwapBufferID):
	      swapBuf = True;
	      break;
            case PEXEncodeEnum(ES_VENDOR_ID, ESEscapeSwapBufferContentID):
	      swapContent = True;
              break;
	    default:
	      break;
	 }
      }
      PEXFreeEnumInfo( 1, enum_counts, enum_desc);
   }
   if ( !(dblBuf && swapBuf && swapContent) ) 
      return( False);

   /* If the back buffer is not cleared on swap then set clear to False */
   swap_content_data.drawable = (Drawable) win;
   swap_content_reply = (esEscapeSwapBufferContentReply *) PEXEscapeWithReply( dpy,
      	PEXEncodeEscape(ES_VENDOR_ID, ESEscapeSwapBufferContentID),
	sizeof(swap_content_data), (char *) &swap_content_data, &reply_length);
   if ( swap_content_reply == NULL ) return False;
   *do_clear = 
      ( swap_content_reply->content != ES_DB_SWAP_CONTENT_CLEAR_TO_BACKGROUND );
   XFree( swap_content_reply );

   /* Turn on double buffering */
   db_data.drawable = (Drawable) win;
   db_data.bufferMode = ES_RENDERER_DBLBUFFER;
   PEXEscape( dpy, PEXEncodeEscape(ES_VENDOR_ID, ESEscapeDoubleBufferID),
     sizeof( db_data), (char *)&db_data);

   return True;
}

int create_pixmap_buffer(dpy, win, rd, ext_info, pixmap_return, width_return, 
                       height_return, depth_return)
Display *dpy;
Window  win;
PEXRenderer rd;
PEXExtensionInfo *ext_info;
Pixmap *pixmap_return;
int *width_return;
int *height_return;
int *depth_return;
{
   XWindowAttributes win_attr;
   PEXRenderingTarget *targets;
   unsigned long count;

   XGetWindowAttributes( dpy, win, &win_attr);
   *width_return = win_attr.width;
   *height_return = win_attr.height;
   *depth_return = win_attr.depth;

   /* Check if pixmaps are supported (this call was introduced in 5.1) */
   if ( ext_info->minor_version > 0 ) {
      PEXMatchRenderingTargets( dpy, win, win_attr.depth, PEXPixmapDrawable, 
	win_attr.visual, MAX_TARGETS, &count, &targets);
      XFree( targets);
      if (count <= 0) return False;
   }

   *pixmap_return = XCreatePixmap( dpy, win, *width_return, *height_return, 
	*depth_return);

   /* On 5.0 servers check for support by trying to use it */
   if ( ext_info->minor_version == 0 ) {
      PEXBeginRendering( dpy, *pixmap_return, rd);
      XSync( dpy, 0);
      if (error_flag != 0) {
          error_flag = 0;
          PEXEndRendering( dpy, rd, False);
          XFreePixmap( dpy, *pixmap_return);
          return False;
      }
      PEXEndRendering( dpy, rd, False);
   }

   return True;
}

int create_back_buffer(dpy, win, rd, db_data, background, bg_pixel)
Display *dpy;
Window  win;
PEXRenderer rd;
bufData *db_data;
PEXColorRGB  *background;
unsigned long bg_pixel;
{
   PEXExtensionInfo *ext_info;
   Multibuffer mbuffers[2];
   Pixmap pixmap;
   int my_err_handler();

   print_err = False;
   XSetErrorHandler( my_err_handler);

   ext_info = PEXGetExtensionInfo( dpy);

   /* Try to create a back buffer using MBX */
   if ( create_MBX_buffer( dpy, win, rd, ext_info, mbuffers) ) {
       db_data->buffer_type = DOUBLE_BUF_MBX;
       db_data->do_clear = False;
       db_data->window = win;
       db_data->front_buffer = (Drawable)mbuffers[0];
       db_data->back_buffer = (Drawable)mbuffers[1];
   } 

   /* Try to create a back buffer using E&S Extension */
   else if ( create_ES_buffer( dpy, win, rd, ext_info, &db_data->do_clear) ) {
       db_data->buffer_type = DOUBLE_BUF_ESCAPE;
       if ( db_data->do_clear )
           db_data->do_clear = !set_clear( dpy, rd, ext_info, background);
       db_data->window = win;
       db_data->front_buffer = 
         db_data->back_buffer = (Drawable) win;
   } 

   /* Try to create a pixmap buffer */
   else if ( create_pixmap_buffer( dpy, win, rd, ext_info, &pixmap, &db_data->width, 
                   &db_data->height, &db_data->depth) ) {
       XGCValues gc_values;
       db_data->buffer_type = DOUBLE_BUF_PIXMAP;
       db_data->do_clear = !set_clear( dpy, rd, ext_info, background);
       db_data->window = win;
       db_data->front_buffer = (Drawable) win;
       db_data->back_buffer = (Drawable) pixmap;
       gc_values.background = bg_pixel;
       db_data->gc = XCreateGC( dpy, win, GCBackground, &gc_values);
   } 

   /* Double buffering isn't supported so set up for single buffering */
   else {
       db_data->buffer_type = SINGLE_BUF;
       db_data->do_clear = !set_clear( dpy, rd, ext_info, background);
       db_data->window = win;
       db_data->front_buffer = 
         db_data->back_buffer = (Drawable) win;
   }

   print_err = True;
   return;
}

int clear_back_buffer(dpy, db_data)
Display *dpy;
bufData *db_data;
{
   /* MBX will clear when it swaps buffers. Also, since we only use E&S      */
   /* double-buffering on 5.1 servers, we know that either the buffer will   */
   /* be cleared at swap or at BeginRendering for this mode. Thus, do_clear  */
   /* is only set when pixmap double buffering or single buffering is chosen */
   /* and we are talking to a 5.0 server */

   if ( db_data->do_clear && db_data->buffer_type == DOUBLE_BUF_PIXMAP )
       XFillRectangle( dpy, db_data->back_buffer, db_data->gc, 0, 0, 
           db_data->width, db_data->height);
   else if ( db_data->do_clear && db_data->buffer_type == SINGLE_BUF )
       XClearWindow( dpy, db_data->window); 
}

int display_back_buffer(dpy, db_data)
Display *dpy;
bufData *db_data;
{

   /* We just just need to swap; MBX will clear the back buffer */
   if ( db_data->buffer_type == DOUBLE_BUF_MBX ) {
       Multibuffer mbuffer = (Multibuffer) db_data->back_buffer;
       Drawable temp;
       XmbufDisplayBuffers( dpy, 1, &mbuffer, 0, 0);
       temp = db_data->back_buffer;
       db_data->back_buffer = db_data->front_buffer;
       db_data->front_buffer = temp;
   } 
   /* E&S double buffering is only used on 5.1 servers; if we get here */
   /* then PEX will clear on swap or at BeginRendering                 */
   else if ( db_data->buffer_type == DOUBLE_BUF_ESCAPE ) {
       esEscapeSwapBuffer swap_data;
       swap_data.drawable = db_data->back_buffer;
       PEXEscape( dpy, PEXEncodeEscape( ES_VENDOR_ID, ESEscapeSwapBufferID),
           sizeof( swap_data), (char *) &swap_data);
   }
   /* For pixmap double-buffering we just copy the pixmap to the window  */
   /* On 5.0 servers, clear_back_buffer needs to be called before drawing*/
   else if ( db_data->buffer_type == DOUBLE_BUF_PIXMAP )
       XCopyArea( dpy, db_data->back_buffer, db_data->front_buffer, db_data->gc, 
	   0, 0, db_data->width, db_data->height, 0, 0);

   /* We don't do anything for single-buffering. On 5.0 servers,  */
   /* clear_back_buffer needs to be called before rendering       */ 

   return;
}

int resize_back_buffer(dpy, width, height, db_data)
Display *dpy;
int width;
int height;
bufData *db_data;
{
   Pixmap pixmap;

   if ( db_data->buffer_type == DOUBLE_BUF_PIXMAP &&
       (db_data->width != width || db_data->height != height) ) {
       XFreePixmap( dpy, (Pixmap)db_data->back_buffer);
       pixmap = XCreatePixmap( dpy, db_data->window, width, height, db_data->depth);
       db_data->width = width;
       db_data->height = height;
       db_data->back_buffer = (Drawable) pixmap;
   }

   return;
}

int destroy_back_buffer(dpy, db_data)
Display *dpy;
bufData *db_data;
{

   if ( db_data->buffer_type == DOUBLE_BUF_MBX ) {
       XmbufDestroyBuffers( dpy, db_data->window);
   }
   else if ( db_data->buffer_type == DOUBLE_BUF_ESCAPE ) {
       esEscapeDblBuffer es_db_data;
       es_db_data.drawable = db_data->back_buffer;
       es_db_data.bufferMode = ES_RENDERER_SINGLEBUFFER;
       PEXEscape( dpy, PEXEncodeEscape( ES_VENDOR_ID, ESEscapeDoubleBufferID),
           sizeof( es_db_data), (char *)&es_db_data);
   }
   else if ( db_data->buffer_type == DOUBLE_BUF_PIXMAP ) {
       XFreeGC( dpy, db_data->gc);
       XFreePixmap( dpy, (Pixmap)db_data->back_buffer);
   }

   db_data->buffer_type = SINGLE_BUF;
   db_data->back_buffer = NULL;

   return;
}
