/*
 *  File Name: color_cube.c
 * 
 *  Author: Paula Womack
 *
 *  Description: Draws color cube which appears in plates
 *
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"

#define	TITLE 		"PEX Color Cube"

#define NUM_VERT 8
static PEXVertexRGBNormal cube_vertices[NUM_VERT] = 
{
    {{ 0.0,  0.0,  1.0}, /* front lower-left vertex */
    { 0.0,  0.0,  1.0}, 
    { 0.0,  0.0,  1.0}}, 
    {{ 1.0,  0.0,  1.0}, /* front lower-right vertex */
    { 1.0,  0.0,  1.0}, 
    { 0.0,  0.0,  1.0}}, 
    {{ 1.0,  1.0,  1.0}, /* front upper-right vertex */
    { 1.0,  1.0,  1.0}, 
    { 0.0,  0.0,  1.0}}, 
    {{ 0.0,  1.0,  1.0}, /* front upper-left vertex */
    { 0.0,  1.0,  1.0}, 
    { 0.0,  0.0,  1.0}}, 
    {{ 0.0,  0.0,  0.0}, /* back lower-left vertex */
    { 0.0,  0.0,  0.0}, 
    { 0.0,  0.0, -1.0}}, 
    {{ 1.0,  0.0,  0.0}, /* back lower-right vertex */
    { 1.0,  0.0,  0.0}, 
    { 0.0,  0.0, -1.0}}, 
    {{ 1.0,  1.0,  0.0}, /* back upper-right vertex */
    { 1.0,  1.0,  0.0}, 
    { 0.0,  0.0, -1.0}}, 
    {{ 0.0,  1.0,  0.0}, /* back upper-left vertex */
    { 0.0,  1.0,  0.0}, 
    { 0.0,  0.0, -1.0}}, 
};
static PEXCoord cube_corner = {0.0, 0.0, 1.0};

#define VERT_PER_FACE   4
#define NUM_FACES	6
unsigned short indices[VERT_PER_FACE * NUM_FACES] =
{
    0, 1, 2, 3,     /* front face */
    1, 5, 6, 2,     /* right face */
    0, 4, 7, 3,     /* left face */
    4, 5, 6, 7,     /* back face */
    3, 2, 6, 7,     /* top face */
    0, 1, 5, 4,     /* bottom face */
};

PEXVector face_data[NUM_FACES] = 
{
    { 0.0,  0.0,  1.0}, /* normal for front face */
    { 1.0,  0.0,  0.0}, /* normal for right face */
    {-1.0,  0.0,  0.0}, /* normal for left face */
    { 0.0,  0.0, -1.0}, /* normal for back face */
    { 1.0,  0.0,  0.0}, /* normal for top face */
    {-1.0,  0.0,  0.0}, /* normal for bottom face */
};

static PEXListOfUShort fillAreas[NUM_FACES];
static PEXConnectivityData fillAreaSets[NUM_FACES];

static PEXColorRGB  black_color = {0.0, 0.0, 0.0};

#define TEXT 		(1)
#define CUBE 		(2)
#define CUBE_INSTANCE 	(3)

int draw_cube ( dpy, rd, scaleMatrix)
Display *dpy; 
PEXRenderer rd; 
PEXMatrix scaleMatrix;
{
    PEXArrayOfFacetData facet_data;
    PEXArrayOfVertex    vertices;

    facet_data.normal = face_data;
    vertices.rgb_normal = cube_vertices;

    PEXBeginStructure(dpy, rd, CUBE);
      PEXSetInteriorStyle( dpy, rd, PEXOCRender, PEXInteriorStyleSolid);
      PEXSetSurfaceInterpMethod( dpy, rd, PEXOCRender, PEXSurfaceInterpColor);

      /* Scale cube so we can fit multiple instances in view */
      PEXSetLocalTransform( dpy, rd, PEXOCRender, PEXReplace, scaleMatrix);
      PEXSetOfFillAreaSets( dpy, rd, PEXOCRender, PEXShapeConvex, 
          PEXGANormal, PEXGAColor | PEXGANormal, PEXGANone, PEXContourDisjoint, 
          True/*one fill area per contour*/, PEXColorTypeRGB,
          NUM_FACES, facet_data, NUM_VERT, vertices, 
          VERT_PER_FACE * NUM_FACES, NULL/*no edge data*/, fillAreaSets);
    PEXEndStructure(dpy, rd);
}

#define PI 	3.1415
int draw_color_cube( dpy, win, rd )
Display *dpy; 
Window win; 
PEXRenderer rd;
{
    int i;
    PEXMatrix scaleMatrix, transMatrix, rotMatrix, xrotMatrix, yrotMatrix;
    PEXVector trans, scale;
    unsigned short *fillAreaIndices = indices;
    PEXRendererAttributes rd_attr;

    /*
     * Initialize the fill areas--there is for each face.
     * Also initialize the fill area sets. Each fill area
     * set just contains one fill area (i.e. one face )
     */
    for ( i = 0; i < NUM_FACES; i++) {
      fillAreas[i].count = VERT_PER_FACE;
      fillAreas[i].shorts = fillAreaIndices;
      fillAreaIndices += VERT_PER_FACE;

      fillAreaSets[i].count = 1; /* just one fill area per set */
      fillAreaSets[i].lists = &fillAreas[i];
    }

    /* 
     * Data ranges from (0,0,0) to (1,1,1). It must be scaled 
     * so we can fit it in the viewing volume
     */
    scale.x = scale.y = scale.z = 0.5;
    PEXScale(&scale, scaleMatrix);

    /* 
     * Get rotation matrix
     */
    PEXRotate (PEXYAxis,  -PI/6.0, yrotMatrix);
    PEXRotate (PEXXAxis, PI/6.0, xrotMatrix);
    PEXMatrixMult (xrotMatrix, yrotMatrix, rotMatrix);

    /* 
     * Must do rotation and scaling about the origin. Thus the composite
     * scale/rotation matrix translates to the origin, does the 
     * scale/rotation, and then translates back
     */
    trans.x = -0.5; trans.y = -0.5; trans.z = -0.5;
    PEXTranslate( &trans, transMatrix);
    PEXMatrixMult (rotMatrix, transMatrix, rotMatrix);
    PEXMatrixMult (scaleMatrix, transMatrix, scaleMatrix);

    trans.x = 0.5; trans.y = 0.5; trans.z = 0.5;
    PEXTranslate( &trans, transMatrix);
    PEXMatrixMult (transMatrix, rotMatrix, rotMatrix);
    PEXMatrixMult (transMatrix, scaleMatrix, scaleMatrix);

    /* Turn on z-buffering */
    rd_attr.hlhsr_mode = PEXHLHSRZBuffer;
    PEXChangeRenderer (dpy, rd, PEXRAHLHSRMode, &rd_attr);

    PEXBeginRendering( dpy, win, rd);
       PEXSetLocalTransform( dpy, rd, PEXOCRender, PEXReplace, rotMatrix);
       draw_cube(dpy, rd, scaleMatrix);

    PEXEndRendering( dpy, rd, True);
}


int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done;   /*indicates when program is done*/

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);

   /* Find best visual and colormap. If we fail, just exit */
    if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE, BlackPixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   /* Ask for mouse press and window exposure events */
   XSelectInput( dpy, win, ButtonPressMask | ExposureMask);

   /* If PEX version 5.1 or later then let PEX clear the window to 
    * black before rendering */
   rd_flags = 0;
   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb.red = 0.0;
      rd_attr.background_color.value.rgb.green = 0.0;
      rd_attr.background_color.value.rgb.blue = 0.0;
      rd_flags |= PEXRAClearImage | PEXRABackgroundColor;
   }

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Event loop */
   done = False;
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window if we are talking to a PEX 5.0 server
          *    Redraw the cube 
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
            if ( ext_info->minor_version < 1 )
               XClearWindow( dpy, win);
            draw_color_cube( dpy, win, rd);
            break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   PEXFreeRenderer( dpy, rd);
   XDestroyWindow( dpy, win);
   XCloseDisplay( dpy);
   exit( 0);
}
