/*
 *  File Name: draw_diagram2.c
 * 
 *  Description: Draws diagrams from Chapter 2
 *
 *  Author: Paula Womack
 *
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"

#define	TITLE 		"Diagram 2.1"

#define N_LINE 5
static PEXCoord diagonal1[N_LINE] =
{
 {-0.26,  -0.26,  -0.3},
 {-0.22,  -0.26,  -0.3},
 { 0.06,   0.06,  -0.3},
 { 0.02,   0.06,  -0.3},
 {-0.26,  -0.26,  -0.3},
};
static PEXCoord diagonal2[N_LINE] =
{
 {-0.26,  0.06,  -0.3},
 {-0.22,  0.06,  -0.3},
 { 0.06, -0.26,  -0.3},
 { 0.02, -0.26,  -0.3},
 {-0.26,  0.06,  -0.3},
};

static PEXCoord xaxis[2] =
{
 {-0.3,   -0.3,  -0.3},
 { 0.3,   -0.3,  -0.3},
};
static PEXCoord yaxis[2] =
{
 {-0.3,   -0.3,  -0.3},
 {-0.3,    0.3,  -0.3},
};
static PEXCoord zaxis[2*3] =
{
 {-0.3,   -0.3,  -0.3},
 {-0.3,   -0.3,   0.3},
};

#define N_PLANE 4
static PEXCoord view_plane[N_PLANE] =
{
 {-0.3,   -0.3,  -0.3},
 { 0.1,   -0.3,  -0.3},
 { 0.1,    0.1,  -0.3},
 {-0.3,    0.1,  -0.3},
};

#define N_FACE 5
static PEXCoord back_face[N_FACE] =
{
 {-0.3,   -0.3,  -0.3},
 { 0.1,   -0.3,  -0.3},
 { 0.1,    0.1,  -0.3},
 {-0.3,    0.1,  -0.3},
 {-0.3,   -0.3,  -0.3},
};
static PEXCoord front_face[N_FACE*3] =
{
 {-0.3,   -0.3,  0.1},
 { 0.1,   -0.3,  0.1},
 { 0.1,    0.1,  0.1},
 {-0.3,    0.1,  0.1},
 {-0.3,   -0.3,  0.1},
};
static PEXCoord connector1[2] =
{
 {-0.3,   -0.3,  -0.3},
 {-0.3,   -0.3,   0.1},
};
static PEXCoord connector2[2] =
{
 { 0.1,   -0.3,  -0.3},
 { 0.1,   -0.3,   0.1},
};
static PEXCoord connector3[2] =
{
 {-0.3,   0.1,  -0.3},
 {-0.3,   0.1,   0.1},
};
static PEXCoord connector4[2] =
{
 { 0.1,   0.1,  -0.3},
 { 0.1,   0.1,   0.1},
};
static PEXColorRGB  black_color = {0.0, 0.0, 0.0};
static PEXColorRGB  grey_color = {0.8, 0.8, 0.8};

int draw_diagram( Display *dpy, Window win, PEXRenderer rd )
{
PEXCoord          origin;
PEXCoord          offset;
PEXMatrix         matrix1, matrix2, rotMatrix, transMatrix;
PEXVector         trans;
PEXCoord          xarrow[4], yarrow[4], zarrow[4];

#define PI 3.1415
    /* Get isometric orientation matrix */
    PEXRotate (PEXYAxis,  PI/4.0, matrix1);
    PEXRotate (PEXXAxis, PI * (30.0/180.0), matrix2);
    PEXMatrixMult (matrix1, matrix2, rotMatrix);

    /* We want to use the origin of the view volume as the origin of the image */
    /* To do this we must:                                                     */
    /*        translate the back left corner to the world coord origin         */
    /*        apply the isometric rotation                                     */
    /*        translate the back left corner back		               */
    trans.x = 0.3; trans.y = 0.3; trans.z = 0.3;
    PEXTranslate( &trans, transMatrix);
    PEXMatrixMult( rotMatrix, transMatrix, matrix1);

    trans.x = -trans.x; trans.y = -trans.y; trans.z = -trans.z;
    trans.y += .2;
    PEXTranslate( &trans, transMatrix);
    PEXMatrixMult( transMatrix, matrix1, rotMatrix);

    /* We need the end of the x, y and z axis. We will use this as the origin */
    /* for our annotation arrows for the axis                                 */
    /* To calculate the arrows assume the projected z-axis portends a         */
    /* 45-degree angle and the y-axis a 60 degree angle from the view plane   */
    PEXTransformPoints (rotMatrix, 1, &xaxis[1], &xarrow[0]);
    PEXTransformPoints (rotMatrix, 1, &yaxis[1], &yarrow[0]);
    PEXTransformPoints (rotMatrix, 1, &zaxis[1], &zarrow[0]);
    xarrow[1] = xarrow[2] = xarrow[3] = xarrow[0]; 
    yarrow[1] = yarrow[2] = yarrow[3] = yarrow[0]; 
    zarrow[1] = zarrow[2] = zarrow[3] = zarrow[0]; 
    xarrow[1].x -= 0.02;  xarrow[1].y += 0.02;
    xarrow[3].x -= 0.02;  xarrow[3].y -= 0.02;
    yarrow[1].y -= 0.025; yarrow[1].x += 0.01;
    yarrow[3].y -= 0.02;  yarrow[3].x -= 0.03;
    zarrow[1].y += 0.035; zarrow[1].x -= 0.005;
    zarrow[3].y += 0.005; zarrow[3].x -= 0.035;

    PEXBeginRendering( dpy, win, rd);
        /* draw the arrows with no rotation .. i.e. annotation */
        PEXSetLineColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
           (PEXColor *)&black_color);
        PEXSetLineWidth( dpy, rd, PEXOCRender, 2.0);
        PEXPolyline( dpy, rd, PEXOCRender, 2, &xarrow[0]);
        PEXPolyline( dpy, rd, PEXOCRender, 2, &xarrow[2]);
        PEXPolyline( dpy, rd, PEXOCRender, 2, &yarrow[0]);
        PEXPolyline( dpy, rd, PEXOCRender, 2, &yarrow[2]);
        PEXPolyline( dpy, rd, PEXOCRender, 2, &zarrow[0]);
        PEXPolyline( dpy, rd, PEXOCRender, 2, &zarrow[2]);

        /* everything else gets transformed */
	PEXSetLocalTransform( dpy, rd, PEXOCRender, PEXReplace, rotMatrix);

        /* draw the axis */
        PEXSetLineWidth( dpy, rd, PEXOCRender, 2.0);
        PEXPolyline( dpy, rd, PEXOCRender, 2, xaxis);
        PEXPolyline( dpy, rd, PEXOCRender, 2, yaxis);
        PEXPolyline( dpy, rd, PEXOCRender, 2, zaxis);

        /* Label the axis -- use nominal line width for the text */
        PEXSetLineWidth( dpy, rd, PEXOCRender, 1.0);
        PEXSetTextColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
            (PEXColor *)&black_color);
        PEXSetATextHeight( dpy, rd, PEXOCRender, 0.022);
	PEXSetTextPrecision( dpy, rd, PEXOCRender, PEXStrokePrecision);

        offset.x = offset.y = offset.z = 0.0;
        origin.x = .32; origin.y = -.3; origin.z = -.3;
        PEXAnnotationText( dpy, rd, PEXOCRender, &origin, &offset, 1, "X");
        origin.x = -.3; origin.y = .32; origin.z = -.3;
        PEXAnnotationText( dpy, rd, PEXOCRender, &origin, &offset, 1, "Y");
        origin.x = -.29; origin.y = -.3; origin.z = .32;
        PEXSetATextHeight( dpy, rd, PEXOCRender, 0.02);
        PEXAnnotationText( dpy, rd, PEXOCRender, &origin, &offset, 1, "Z");

        /* Label the view volume */
        origin.x = -.5; origin.y = -.3; origin.z = -.3;
        PEXAnnotationText( dpy, rd, PEXOCRender, &origin, &offset, 7, "(0,0,0)");
        origin.x = .12; origin.y = .1; origin.z = .1;
        PEXAnnotationText( dpy, rd, PEXOCRender, &origin, &offset, 7, "(1,1,1)");
         
        /* draw shaded surface with X in it to illustrate view plane */
        PEXSetLineWidth( dpy, rd, PEXOCRender, 2.0);
        PEXSetSurfaceColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
            (PEXColor *)&grey_color);
	PEXSetInteriorStyle ( dpy, rd, PEXOCRender, PEXInteriorStyleSolid);
	PEXFillArea( dpy, rd, PEXOCRender, PEXShapeConvex, False, N_PLANE, view_plane);

        PEXPolyline( dpy, rd, PEXOCRender, N_LINE, diagonal1);
        PEXPolyline( dpy, rd, PEXOCRender, N_LINE, diagonal2);

        /* draw the view volume as a dashed unit cube */
        PEXSetLineType( dpy, rd, PEXOCRender, PEXLineTypeDashed);
	PEXPolyline( dpy, rd, PEXOCRender, N_FACE, front_face);
	PEXPolyline( dpy, rd, PEXOCRender, N_FACE, back_face);
        PEXPolyline( dpy, rd, PEXOCRender, 2, connector1);
        PEXPolyline( dpy, rd, PEXOCRender, 2, connector2);
        PEXPolyline( dpy, rd, PEXOCRender, 2, connector3);
        PEXPolyline( dpy, rd, PEXOCRender, 2, connector4);
    PEXEndRendering( dpy, rd, True);
}


int main(int argc, char *argv[])
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXViewEntry         	view_entry;
   PEXCoord          		from;
   PEXCoord          		to;
   PEXVector         		up;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done;   /*indicates when program is done*/

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);

   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE, WhitePixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   /* Ask for mouse press and window exposure events */
   XSelectInput( dpy, win, ButtonPressMask | ExposureMask);

   /* Create a PEX view table so we can center view around origin */
   from.x = 0.0;
   from.y = 0.0;
   from.z = 9.0;
   to.x = 0.0;
   to.y = 0.0;
   to.z = 0.0;
   up.x = 0.0;
   up.y = 1.0;
   up.z = 0.0;
   PEXLookAtViewMatrix (&from, &to, &up, view_entry.orientation);
   PEXOrthoProjMatrix (1.3, 1.0, 2.0, -2.0, view_entry.mapping );
   view_entry.clip_flags = 0;

   rd_attr.view_table = PEXCreateLookupTable( dpy, win, PEXLUTView);
   PEXSetTableEntries( dpy, rd_attr.view_table, 0/*default entry*/,
	        1/*count*/,  PEXLUTView, &view_entry);
   rd_flags = PEXRAViewTable;

   /* If PEX version 5.1 or later then let PEX clear the window to
    * white before rendering */
   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb.red = 1.0;
      rd_attr.background_color.value.rgb.green = 1.0;
      rd_attr.background_color.value.rgb.blue = 1.0;
      rd_flags |= PEXRAClearImage | PEXRABackgroundColor;
   }

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Event loop */
   done = False;
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window if we are talking to a PEX 5.0 server
          *    Redraw the diagram
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
            if ( ext_info->minor_version < 1 )
               XClearWindow( dpy, win);
            draw_diagram( dpy, win, rd);
            break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   PEXFreeRenderer( dpy, rd);
   XDestroyWindow( dpy, win);
   XCloseDisplay( dpy);
   exit( 0);
}
