/*
 *  File Name: animated_teapot.c
 *
 *  Description: Animated teapot example from Chapter 12
 * 
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "view_utils.h"
#include "util_protos.h"
#include "teapot_data.h"

#define	TITLE 		"PEX Teapot"

#define MAX_LIGHTS 10 

static PEXColorRGB  red_color = {1.0, 0.0, 0.0};
static PEXColorRGB  background_color = {0.0, 0.0, 0.0};

/* 
 * Create some lights
 */
int add_lights( dpy, win, light_table, lights)
Display *dpy; 
Window win; 
PEXLookupTable light_table;
PEXTableIndex *lights;

{
    unsigned short imp_dep_const_name;
    PEXImpDepConstant *imp_dep_const;
    int num_non_ambients;
    int enum_type_name;
    unsigned long *enum_counts;
    PEXEnumTypeDesc *enum_desc;
    int num_enums;
    int do_ambient, do_vector, do_point;
    PEXLightEntry light_entry;
    PEXVector light_dir;
    int nlights = 0;

    /* Find out how many non-ambient lights can be enabled at once */
    imp_dep_const_name = PEXIDMaxNonAmbientLights;
    if (!PEXGetImpDepConstants (dpy, win, 1, &imp_dep_const_name, &imp_dep_const) ) {
        fprintf( stderr, "Error calling PEXGetImpDepConstants\n");
        return nlights;
    }
    num_non_ambients = imp_dep_const[0].integer;
    XFree( imp_dep_const);

    /* 
     * Get the supported light types. Start with all light types disabled then loop
     * through the list of supported lights enabling the types that are supported
     */
    enum_type_name = PEXETLightType;
    if (!PEXGetEnumTypeInfo( dpy, win, 1, &enum_type_name, PEXETIndex, 
        &enum_counts, &enum_desc) ) {
        fprintf( stderr, "Error calling PEXGetEnumTypeInfo\n");
        return nlights;
    }

    do_ambient = do_vector = do_point = False;
    num_enums = enum_counts[0];
    while (--num_enums >= 0) {
        if (enum_desc[num_enums].index == PEXLightAmbient) do_ambient = True;
        if (enum_desc[num_enums].index == PEXLightWCVector) do_vector = True;
        if (enum_desc[num_enums].index == PEXLightWCPoint) do_point = True;
    }
    PEXFreeEnumInfo( 1, enum_counts, enum_desc);

    /* If ambient lights are supported then add a gray ambient light */
    if (do_ambient) {
        light_entry.type = PEXLightAmbient;
        light_entry.color.type = PEXColorTypeRGB;
        light_entry.color.value.rgb.red = 0.4;
        light_entry.color.value.rgb.green = 0.4;
        light_entry.color.value.rgb.blue = 0.4;
        PEXSetTableEntries( dpy, light_table, 1/*entry*/, 1/*count*/,
            PEXLUTLight, &light_entry);
        lights[nlights++] = 1;
    }

    /* 
     * If vector lights are supported and we haven't exceeded the maximum 
     * non-ambients then add a white vector light coming from the front left 
     */ 
    if (do_vector && num_non_ambients > 0) {
        light_entry.type = PEXLightWCVector;
        light_dir.x =  1.0;
        light_dir.y =  0.0;
        light_dir.z = -1.0;
        PEXNormalizeVectors( 1/*count*/, &light_dir, &light_entry.direction);
        light_entry.color.type = PEXColorTypeRGB;
        light_entry.color.value.rgb.red = 0.6;
        light_entry.color.value.rgb.green = 0.6;
        light_entry.color.value.rgb.blue = 0.6;
        PEXSetTableEntries( dpy, light_table, 2/*entry*/, 1/*count*/,
            PEXLUTLight, &light_entry);
        lights[nlights++] = 2;
        num_non_ambients--;
    }

    /* 
     * If point lights are supported and we haven't exceeded the maximum 
     * non-ambients then add a white point light overhead
     */ 
    if (do_point && num_non_ambients > 0) {
        light_entry.type = PEXLightWCPoint;
        light_entry.point.x = (PointMaxX - PointMinX)/2.0 + PointMinX;
        light_entry.point.y = PointMaxY + (PointMaxY - PointMinY)/10.0;
        light_entry.point.z = (PointMaxZ - PointMinZ)/2.0 + PointMinZ;
        light_entry.color.type = PEXColorTypeRGB;
        light_entry.color.value.rgb.red = 0.6;
        light_entry.color.value.rgb.green = 0.6;
        light_entry.color.value.rgb.blue = 0.6;
        light_entry.attenuation1 = 1.0;
        light_entry.attenuation2 = 5.0;
        PEXSetTableEntries( dpy, light_table, 3/*entry*/, 1/*count*/,
            PEXLUTLight, &light_entry);
        lights[nlights++] = 3;
        num_non_ambients--;
    }

    return nlights;
}

PEXStructure store_teapot( dpy, num_lights, lights)
Display *dpy;
int num_lights;
PEXTableIndex *lights;
{
    extern PEXVector fnormals[];
    extern PEXVector vnormals[];
    extern PEXCoord  vpoints[];

    int i, j, k;
    PEXArrayOfVertex  vertices;
    PEXFacetData      facet_data;
    PEXVertexNormal   vertex[3];
    PEXStructure sid;
    PEXReflectionAttributes reflection_attr;

    sid = PEXCreateStructure( dpy);
 
    PEXSetSurfaceColor( dpy, sid, PEXOCStore, PEXColorTypeRGB, 
        (PEXColor *)&red_color);
    PEXSetLightSourceState( dpy, sid, PEXOCStore, num_lights, lights, 0, NULL);
    PEXSetInteriorStyle( dpy, sid, PEXOCStore, PEXInteriorStyleSolid);
    PEXSetSurfaceInterpMethod( dpy, sid, PEXOCStore, PEXSurfaceInterpColor);
    PEXSetReflectionModel( dpy, sid, PEXOCStore, PEXReflectionSpecular);

    reflection_attr.ambient = 0.8;
    reflection_attr.diffuse = 0.8;
    reflection_attr.specular = 1.0;
    reflection_attr.specular_conc = 10.0;
    reflection_attr.transmission = 0.0;
    reflection_attr.specular_color.type = PEXColorTypeRGB;
    reflection_attr.specular_color.value.rgb.red = 1.0;
    reflection_attr.specular_color.value.rgb.green = 1.0;
    reflection_attr.specular_color.value.rgb.blue = 1.0;
    PEXSetReflectionAttributes( dpy, sid, PEXOCStore, &reflection_attr);
        
    PEXSetFacetCullingMode( dpy, sid, PEXOCStore, PEXBackFaces);

    vertices.normal = vertex;
    for ( i = 0, j = 0; i < NumFacets; i++, j+=3 ) 
    {
	facet_data.normal = fnormals[i];
	for (k = 0; k < 3; k++)
	{
	    vertex[k].normal = vnormals[j+k];
	    vertex[k].point  = vpoints[j+k];
	}
	PEXFillAreaWithData( dpy, sid, PEXOCStore, PEXShapeConvex, False, 
		PEXGANormal, PEXGANormal, 0/*no colors*/, 
		&facet_data, 3, vertices); 
    }

    return sid;
}

int draw_teapot (dpy, draw, rd, str, rot_matrix)
Display *dpy;
Drawable draw;
PEXRenderer rd;
PEXStructure str;
PEXMatrix rot_matrix;
{
    PEXBeginRendering( dpy, draw, rd);
      PEXSetLocalTransform( dpy, rd, PEXOCRender, PEXReplace, rot_matrix);
      PEXExecuteStructure( dpy, rd, PEXOCRender, str);
    PEXEndRendering( dpy, rd, True);

    return;
}

int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   PEXStructure                 sid;
   int 				num_lights;
   PEXTableIndex 	        lights[MAX_LIGHTS];
   PEXMatrix                    rot_matrix, tmp_rot_matrix;
   int				done;   /*indicates when program is done*/
   int			        redraw, button_press, compute_rotation, rotate;
   char 			key_pressed;
   int			        firstx, firsty, lastx, lasty;
   float			dx, dy, angx, angy;
   bufData			db_data;
   PEXCoord			teapot_center;
   PEXVector			trans_vector, scale_vector;
   int				subset;

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);
    
   /* Make sure the structure subset is supported */
   subset = ext_info->subset_info & 0xff;
   if ( (subset != 0) && !(subset & PEXStructureMode) ) {
      fprintf( stderr, "Structures are not supported\n");
      exit( 1);
   }

   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE, BlackPixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);
  
   /* Ask for mouse press and window exposure events */
   XSelectInput( dpy, win, ExposureMask | ButtonPressMask | ButtonMotionMask | 
        KeyPressMask | StructureNotifyMask);

   /* Create PEX view such that object is in center */
   rd_attr.view_table = PEXCreateLookupTable( dpy, win, PEXLUTView);
   rd_flags = PEXRAViewTable;
   init_object_view( dpy, PointMinX, PointMaxX, PointMinY, PointMaxY, PointMinZ,
      PointMaxZ, 1.0/*scale*/, ORTHOGRAPHIC_PROJECTION, FRONT_VIEW, 
      0/*default entry*/, rd_attr.view_table);

   /* Enable Z-buffering */
   rd_attr.hlhsr_mode = PEXHLHSRZBuffer;
   rd_flags |= PEXRAHLHSRMode;

   /* Add some lights */
   rd_attr.light_table = PEXCreateLookupTable( dpy, win, PEXLUTLight);
   rd_flags |= PEXRALightTable;
   num_lights = add_lights( dpy, win, rd_attr.light_table, lights);

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Set up for double buffering */
   create_back_buffer( dpy, win, rd, &db_data, &background_color, 
	BlackPixel( dpy, DefaultScreen( dpy)));

   /* Map the window */
   XMapWindow( dpy, win);
  
   /* Use identity matrix initially */
   teapot_center.x = PointMinX + (PointMaxX - PointMinX)/2;
   teapot_center.y = PointMinY + (PointMaxY - PointMinY)/2;
   teapot_center.z = PointMinZ + (PointMaxZ - PointMinZ)/2;
   PEXIdentityMatrix( rot_matrix);

   /* Store teapot */
   sid = store_teapot( dpy, num_lights, lights);

   /* Event loop */
   done = redraw = button_press = compute_rotation = rotate = False;
   trans_vector.x = trans_vector.y = trans_vector.z = 0.0;
   scale_vector.x = scale_vector.y = scale_vector.z = 1.0;
   while ( !done ) {

      if (XPending(dpy)) {
          XNextEvent( dpy, &event);
          switch ( event.type ) {
             /* Enable drawing as soon as expose event received */
             case Expose:
                redraw = True;
                break;

             /* Record window width and height because it might have changed */
             case ConfigureNotify:
                width = event.xconfigure.width;
                height = event.xconfigure.height;
                resize_back_buffer( dpy, width, height, &db_data);
                redraw = True;
                break;

             /* Note first position on button press, also stop rotation */
             case ButtonPress:
                firstx = event.xbutton.x;
                firsty = event.xbutton.y;
                button_press = True;
                rotate = False;
                break;

             case MotionNotify:
                lastx = event.xmotion.x;
                lasty = event.xmotion.y;
                if (button_press) compute_rotation = True;
                break;
    
             /* Quit when 'q' or 'Q' is typed */
	     case KeyPress:
                XLookupString( (XKeyEvent *)&event, &key_pressed, 1, NULL, NULL);
                if ( key_pressed == 'q' || key_pressed == 'Q' ) {
                    done = True;
                    redraw = rotate = False;
                }
                else if ( key_pressed == 's' || key_pressed == 'S' ) {
                    redraw = rotate = False;
                }
                else if ( key_pressed == 'r' || key_pressed == 'R' ) {
                   PEXIdentityMatrix( rot_matrix);
                   redraw = True;
                }
	        break;
          }

          continue;
      }

      if ( redraw ) {
         if (db_data.do_clear) clear_back_buffer( dpy, &db_data);
         draw_teapot( dpy, db_data.back_buffer, rd, sid, rot_matrix);
         display_back_buffer( dpy, &db_data);
         redraw = False;
      }

      if ( compute_rotation ) {
          dx = lastx - firstx;
          dx = dx / width;
          angy = dx/M_PI;

          dy = firsty - lasty;
          dy = dy / height;
          angx = - dy/M_PI;

          button_press = compute_rotation = False;
          rotate = True;
      }

      if ( rotate ) {
         int i = 10;
         PEXBuildTransform( &teapot_center, &trans_vector, angx, angy, 0.0,
            &scale_vector, tmp_rot_matrix);
         while (i--) {
            PEXMatrixMult( tmp_rot_matrix, rot_matrix, rot_matrix);

            if (db_data.do_clear) clear_back_buffer( dpy, &db_data);
            draw_teapot( dpy, db_data.back_buffer, rd, sid, rot_matrix);
            display_back_buffer( dpy, &db_data);
         }

         XSync( dpy, 0);
      }

   }

   /* Termination */
   destroy_back_buffer( dpy, &db_data);
   XCloseDisplay( dpy);
   exit( 0);
}
