/*
 *  File Name: bspline_curve.c
 *
 *  Description: B-Spline curve example from Chapter 5
 * 
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"

#define	TITLE 		"PEX B-Spline Curve"
#define N_VERT 7

/* 
 * Used to draw b-spline curve 
 */
static PEXCoord curve_control_pts[N_VERT] =
{
 {0.05,  0.4,   0.0},
 {0.2,   0.55,  0.0},
 {0.35,  0.4,   0.0},
 {0.5,   0.55,  0.0},
 {0.65,  0.4,   0.0},
 {0.8,   0.55,  0.0},
 {0.95,  0.4,   0.0},
};

static PEXColorRGB  red_color = {1.0, 0.0, 0.0};

int draw_bspline_curve ( dpy, win, rd )
Display *dpy; 
Window win; 
PEXRenderer rd;
{
    float knots[2*N_VERT];
    double tmin, tmax;
    PEXArrayOfCoord control_pts;

    /* 
     * 7 control points and order 4 requires 11 knots
     * The first and last knot values must be repeated 4 times
     */
    knots[0] = knots[1] = knots[2] = knots[3] = 0.0;
    knots[4] = .25;
    knots[5] = .50;
    knots[6] = .75;
    knots[7] = knots[8] = knots[9] = knots[10] = 1.0;
 
    /* Just set  tmin and tmax to min and max knot values */
    tmin = 0.0;
    tmax = 1.0;
    control_pts.point = curve_control_pts;

    PEXBeginRendering( dpy, win, rd);
        /* Use markers to show control points */
	PEXSetMarkerColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
	    (PEXColor *)&red_color);
        PEXMarkers( dpy, rd, PEXOCRender, N_VERT, curve_control_pts);

        /* Draw curve with order 4 */
	PEXSetLineWidth( dpy, rd, PEXOCRender, 2.0);
	PEXSetLineColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
	    (PEXColor *)&red_color);
        PEXNURBCurve( dpy, rd, PEXOCRender, PEXNonRational, 4, knots, 
		N_VERT, control_pts, tmin, tmax);
    PEXEndRendering( dpy, rd, True);
}

int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done;   /*indicates when program is done*/

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);
 
   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE, WhitePixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   /* Ask for mouse press and window exposure events */
   XSelectInput( dpy, win, ButtonPressMask | ExposureMask);

   /* If PEX version 5.1 or later then let PEX clear the window to 
    * white before rendering */
   rd_flags = 0;
   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb.red = 1.0;
      rd_attr.background_color.value.rgb.green = 1.0;
      rd_attr.background_color.value.rgb.blue = 1.0;
      rd_flags |= PEXRAClearImage | PEXRABackgroundColor;
   }

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Event loop */
   done = False;
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window if we are talking to a PEX 5.0 server
          *    Redraw the curve
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
            if ( ext_info->minor_version < 1 )
               XClearWindow( dpy, win);
            draw_bspline_curve( dpy, win, rd);
            break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   XCloseDisplay( dpy);
   exit( 0);
}
