/*
 *  File Name: bspline_surface.c
 * 
 *  Description: B-spline surface example from Chapter 7
 *
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"

#define	TITLE 		"PEX B-Spline Surface"

/*
 * Used to draw b-spline surface
 */
#define N_SVERT 5
static PEXCoord surf_control_pts[N_SVERT * N_SVERT] =
{
 {0.1,  0.7,   0.7},
 {0.3,  0.85,  0.6},
 {0.5,  0.7,   0.5},
 {0.7,  0.85,  0.6},
 {0.9,  0.7,   0.7},

 {0.1,  0.55,  0.3},
 {0.3,  0.7,   0.4},
 {0.5,  0.55,  0.5},
 {0.7,  0.7,   0.4},
 {0.9,  0.55,  0.3},

 {0.1,  0.4,   0.7},
 {0.3,  0.55,  0.6},
 {0.5,  0.4,   0.5},
 {0.7,  0.55,  0.6},
 {0.9,  0.4,   0.7},

 {0.1,  0.25,  0.3},
 {0.3,  0.4,   0.4},
 {0.5,  0.25,  0.5},
 {0.7,  0.4,   0.4},
 {0.9,  0.25,  0.3},

 {0.1,  0.1,   0.7},
 {0.3,  0.25,  0.6},
 {0.5,  0.1,   0.5},
 {0.7,  0.25,  0.6},
 {0.9,  0.1,   0.7},
};

/*
 * Specify 4 trimming curves to form a square outer loop
 * This determines what part of the surface is kept
 * so it must be counterclockwise
 */
static PEXCoord2D outer_cpt1[2] =
{
 0.0, 0.0,
 1.0, 0.0,
};
static PEXCoord2D outer_cpt2[2] =
{
 1.0, 0.0,
 1.0, 1.0,
};
static PEXCoord2D outer_cpt3[2] =
{
 1.0, 1.0,
 0.0, 1.0,
};
static PEXCoord2D outer_cpt4[2] =
{
 0.0, 1.0,
 0.0, 0.0,
};

/*
 * Specify 4 trimming curves to form a square inner loop
 * This defines the hole so it must be a clockwise loop.
 */
static PEXCoord2D inner_cpt1[2] =
{
 0.7, 0.3,
 0.3, 0.3,
};
static PEXCoord2D inner_cpt2[2] =
{
 0.3, 0.3,
 0.3, 0.7,
};
static PEXCoord2D inner_cpt3[2] =
{
 0.3, 0.7,
 0.7, 0.7,
};
static PEXCoord2D inner_cpt4[2] =
{
 0.7, 0.7,
 0.7, 0.3,
};

static PEXColorRGB  red_color = {1.0, 0.0, 0.0};
static PEXColorRGB  black_color = {0.0, 0.0, 0.0};

int draw_bspline_surf( dpy, win, rd, trim )
Display *dpy; 
Window win; 
PEXRenderer rd; 
int trim;
{
#define N_TVERT 4
    float sknots[2*N_SVERT], tknots[2*N_TVERT];
    PEXTrimCurve tcurve[8];
    PEXListOfTrimCurve tloop[2];
    PEXPSCIsoparametricCurves iso_data;
    int i;
    PEXArrayOfCoord control_pts;

    PEXBeginRendering( dpy, win, rd);

	PEXSetSurfaceColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
	    (PEXColor *)&red_color);
	PEXSetInteriorStyle( dpy, rd, PEXOCRender, PEXInteriorStyleSolid);

        /* 
         * Use uniform placement to draw 5 isoparametric curves on the 
         * surface in both the u and v directions. The curves will be 
         * drawn with the current line attributes
         */
        iso_data.placement_type = PEXUniformPlacement;
        iso_data.u_count = 5;
        iso_data.v_count = 5;
        PEXSetParaSurfCharacteristics( dpy, rd, PEXOCRender, PEXPSCIsoCurves, 
		(PEXPSCData *)&iso_data);
	PEXSetLineColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
	    (PEXColor *)&black_color);
	PEXSetLineWidth( dpy, rd, PEXOCRender, 2.0);

        /* 
         * 5 control points and order 3 requires 8 knots
         * The first and last knot values must be repeated 3 times
         */
        sknots[0] = sknots[1] = sknots[2] = 0.0;
        sknots[3] = 0.33;
        sknots[4] = 0.66;
        sknots[5] = sknots[6] = sknots[7] = 1.0;

        /* 
         * The trim curves have 2 control points and order 2, requiring 4 knots
         */
        tknots[0] = tknots[1] = 0.0; 
        tknots[2] = tknots[3] = 1.0;

        /*
         * There are 2 trimming loops each containing 4 b-splines which 
         * define a square trim loop. 
         */
        tloop[0].count = tloop[1].count = 4;
        tloop[0].curves = &tcurve[0];
        tloop[1].curves = &tcurve[4];

        /* Initialize the trim curves */
        for (i = 0; i < 8; i++) {
            /* When surface edges are on the trim curves will be drawn */
	    tcurve[i].visibility = PEXOn;

            /* Use order 2 so curve goes through all control points */
	    tcurve[i].order = 2;

            /* Don't need homogenous coordinates */
	    tcurve[i].rationality = PEXNonRational;

            /* Tolerance is not used for this approximation */
	    tcurve[i].approx_method = PEXApproxImpDep;

            /* Evaluate curve from min knot value to max knot value */
	    tcurve[i].tmin = tknots[0];
	    tcurve[i].tmax = tknots[3];
	    tcurve[i].knots.count = 4;
	    tcurve[i].knots.floats = tknots;

            /* All the trim curves have 2 points */
	    tcurve[i].count = 2;
	}

        /* First trimming loop determines which part to keep */
	tcurve[0].control_points.point_2d = outer_cpt1;
	tcurve[1].control_points.point_2d = outer_cpt2;
	tcurve[2].control_points.point_2d = outer_cpt3;
	tcurve[3].control_points.point_2d = outer_cpt4;

        /* Second trimming loop defines hole */
	tcurve[4].control_points.point_2d = inner_cpt1;
	tcurve[5].control_points.point_2d = inner_cpt2;
	tcurve[6].control_points.point_2d = inner_cpt3;
	tcurve[7].control_points.point_2d = inner_cpt4;

        control_pts.point = surf_control_pts;
        if (trim)  {
           PEXNURBSurface( dpy, rd, PEXOCRender, PEXNonRational, 3, 3, sknots, 
		sknots, N_SVERT, N_SVERT, control_pts, 2, tloop);
        } else {
           PEXNURBSurface( dpy, rd, PEXOCRender, PEXNonRational, 3, 3, sknots, 
		sknots, N_SVERT, N_SVERT, control_pts, 0, NULL);
        }

    PEXEndRendering( dpy, rd, True);
}

int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done;   /*indicates when program is done*/
   char				key_pressed;
   int 				trim;

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);
 
   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE, WhitePixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   /* Ask for key press, mouse press and window exposure events */
   XSelectInput( dpy, win, KeyPressMask | ButtonPressMask | ExposureMask);

   /* If PEX version 5.1 or later then let PEX clear the window to 
    * white before rendering */
   rd_flags = 0;
   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb.red = 1.0;
      rd_attr.background_color.value.rgb.green = 1.0;
      rd_attr.background_color.value.rgb.blue = 1.0;
      rd_flags |= PEXRAClearImage | PEXRABackgroundColor;
   }

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Event loop */
   done = False;
   trim = False;
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window if we are talking to a PEX 5.0 server
          *    Redraw the b-spline surface
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
            if ( ext_info->minor_version < 1 )
               XClearWindow( dpy, win);
            draw_bspline_surf( dpy, win, rd, trim);
            break;

         /* Alternate between trimmed and untrimmed surface with key press */
         case KeyPress:
            XLookupString((XKeyEvent *)&event, &key_pressed, 1, NULL, NULL);
            if (key_pressed == 't' || key_pressed == 'T') {
	        trim = !trim;
                if ( ext_info->minor_version < 1 )
                   XClearWindow( dpy, win);
                draw_bspline_surf( dpy, win, rd, trim);
            }
            break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   XCloseDisplay( dpy);
   exit( 0);
}
