/*
 *  File Name: cube_ns.c
 *
 *  Description: Nameset example from Chapter 13
 * 
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"

#define	TITLE 		"PEX Cube Instances"

#define MINX 0.0
#define MINY 0.0
#define MINZ 0.0
#define MAXX 1.0
#define MAXY 1.0
#define MAXZ 1.0
#define N_CUBE_VERT 8
static PEXCoord cube_vertices[N_CUBE_VERT] =
{
    {MINX, MINY, MAXZ}, /* front vertices */
    {MAXX, MINY, MAXZ},
    {MAXX, MAXY, MAXZ},
    {MINX, MAXY, MAXZ},
    {MINX, MINY, MINZ}, /* back vertices */
    {MAXX, MINY, MINZ},
    {MAXX, MAXY, MINZ},
    {MINX, MAXY, MINZ},
};
static PEXCoord cube_corner = {0.0, 0.0, 0.0};
static PEXCoord cube_center = {0.5, 0.5, 0.5};

#define VERT_PER_FACE   4
#define NUM_FACES	6
unsigned short indices[VERT_PER_FACE * NUM_FACES] =
{
    0, 1, 2, 3,     /* front face */
    1, 5, 6, 2,     /* right face */
    0, 4, 7, 3,     /* left face */
    4, 5, 6, 7,     /* back face */
    3, 2, 6, 7,     /* top face */
    0, 1, 5, 4,     /* bottom face */
};

PEXColorRGBNormal face_data[NUM_FACES] =
{/* color	               normal */
 {{1.0, 0.0, 0.0},/*red*/     { 0.0, 0.0,  1.0}}, /* front face */
 {{0.0, 1.0, 0.0},/*green*/   { 1.0, 0.0,  0.0}}, /* right face */
 {{0.0, 0.0, 1.0},/*blue*/    {-1.0, 0.0,  0.0}}, /* left face */
 {{1.0, 1.0, 0.0},/*yellow*/  { 0.0, 0.0, -1.0}}, /* back face */
 {{1.0, 0.0, 1.0},/*magenta*/ { 1.0, 0.0,  0.0}}, /* top face */
 {{0.0, 1.0, 1.0},/*cyan*/    {-1.0, 0.0,  0.0}}, /* bottom face */
};

static PEXColorRGB  black_color = {0.0, 0.0, 0.0};
static PEXColorRGB  red_color = {1.0, 0.0, 0.0};

static PEXStructure cube, cube1, cube2, cube3, cube4;

#define NUM_CUBES 4
PEXName cube_names[NUM_CUBES] = {1, 2, 3, 4};

PEXStructure store_cube( dpy)
Display *dpy; 
{
    PEXStructure sid;
    int  i;
    PEXListOfUShort fill_areas[NUM_FACES];
    PEXConnectivityData fill_area_sets[NUM_FACES];
    unsigned short *fill_area_indices = indices;
    PEXArrayOfFacetData facets;
    PEXArrayOfVertex vertices;

    /*
     * Initialize the fill areas--there is one for each face.
     * Also initialize the fill area sets. Each fill area
     * set just contains one fill area (i.e. one face )
     */
    for ( i = 0; i < NUM_FACES; i++) {
      fill_areas[i].count = VERT_PER_FACE;
      fill_areas[i].shorts = fill_area_indices;
      fill_area_indices += VERT_PER_FACE;

      fill_area_sets[i].count = 1; /* just one fill area per set */
      fill_area_sets[i].lists = &fill_areas[i];
    }
    facets.rgb_normal = face_data;
    vertices.no_data = cube_vertices;

    sid = PEXCreateStructure( dpy);

    /* Now store the surface data in the structure */
    PEXSetOfFillAreaSets( dpy, sid, PEXOCStore, PEXShapeConvex, 
        PEXGAColor | PEXGANormal, PEXGANone, PEXGANone, PEXContourDisjoint, 
        True/*one fill area per contour*/, PEXColorTypeRGB,
        NUM_FACES, facets, N_CUBE_VERT, vertices, 
        VERT_PER_FACE * NUM_FACES, NULL/*no edge data*/, fill_area_sets);

    return sid;
}

PEXStructure store_text( dpy, str)
Display *dpy; 
char *str;
{
    PEXStructure sid;
    PEXCoord offset;

    /* offset text from lower left corner of cubes */
    offset.x = offset.y = -.10; offset.z = 0.0;

    sid = PEXCreateStructure( dpy);

    PEXAnnotationText( dpy, sid, PEXOCStore, &cube_corner, &offset,
        strlen(str), str);

    return sid;
}

PEXStructure store_cube_instance( dpy, scale_matrix, rot_matrix, cube, text, name)
Display *dpy; 
PEXMatrix scale_matrix;
PEXMatrix rot_matrix;
PEXStructure cube; 
PEXStructure text;
PEXName name;
{
    PEXStructure sid;

    sid = PEXCreateStructure( dpy);

    /* Scale data so it fits in view volume */
    PEXSetLocalTransform( dpy, sid, PEXOCStore, PEXReplace, scale_matrix);

    /* Set text attributes (to draw red text) and execute text */
    PEXSetTextColor( dpy, sid, PEXOCStore, PEXColorTypeRGB, 
        (PEXColor *) &red_color);
    PEXSetATextHeight( dpy, sid, PEXOCStore, .020);
    PEXSetTextPrecision( dpy, sid, PEXOCStore, PEXStrokePrecision);
    PEXExecuteStructure( dpy, sid, PEXOCStore, text);

    /* Set rotation for cube .. must postconcatentate so scale is done first */
    if (rot_matrix != NULL)
       PEXSetLocalTransform( dpy, sid, PEXOCStore, PEXPostConcatenate, rot_matrix);

    /* Set the surface attributes and then execute the cube */
    PEXSetInteriorStyle( dpy, sid, PEXOCStore, PEXInteriorStyleSolid);
    PEXSetSurfaceEdgeFlag( dpy, sid, PEXOCStore, PEXOn);
    PEXSetSurfaceEdgeColor( dpy, sid, PEXOCStore, PEXColorTypeRGB,
        (PEXColor *) &black_color);
    PEXSetSurfaceEdgeWidth( dpy, sid, PEXOCStore, 2.0);

    PEXAddToNameSet( dpy, sid, PEXOCStore, 1, &name);
    PEXExecuteStructure( dpy, sid, PEXOCStore, cube);
    PEXRemoveFromNameSet( dpy, sid, PEXOCStore, 1, &name);

    return sid;
}

PEXStructure store_cubes( dpy)
Display *dpy;
{
    PEXMatrix scale_matrix, trans_matrix, rot_matrix, xrot_matrix, yrot_matrix;
    PEXVector trans_vector, scale_vector;
    PEXCoord center;
    PEXStructure text1, text2, text3, text4;
    PEXStructure root;

    /* 
     * Data ranges from (0,0,0) to (1,1,1). It must be scaled 
     * so we can fit four cubes in the viewing volume. We use
     * BuildTransform so we can scale about center of cube
     */
    scale_vector.x = scale_vector.y = scale_vector.z = 0.2;
    trans_vector.x = trans_vector.y = trans_vector.z = 0.0;
    PEXBuildTransform (&cube_center, &trans_vector, 0.0, 0.0, 0.0,
        &scale_vector, scale_matrix);

    /* 
     * Get three different rotation matrices --one to rotate about X,
     * one about Y and one about X and Y. Rotations need to be about
     * cube center.
     */
    scale_vector.x = scale_vector.y = scale_vector.z = 1.0;
    PEXBuildTransform( &cube_center, &trans_vector, M_PI/6.0, 0.0, 0.0,
        &scale_vector, xrot_matrix);
    PEXBuildTransform( &cube_center, &trans_vector, 0.0, M_PI/6.0, 0.0,
        &scale_vector, yrot_matrix);
    PEXBuildTransform( &cube_center, &trans_vector, M_PI/6.0, M_PI/6.0, 0.0,
        &scale_vector, rot_matrix);

    /* Put cube in structure so we can reference it multiple times */
    cube = store_cube( dpy);

    /* Now create cubes with different orientations and label them */
    text1 = store_text( dpy, "No Rotation");
    cube1 = store_cube_instance( dpy, scale_matrix, NULL, cube, text1, 
	cube_names[0]);
    text2 = store_text( dpy, "Rotate About X");
    cube2 = store_cube_instance( dpy, scale_matrix, xrot_matrix, cube, text2, 
	cube_names[1]);
    text3 = store_text( dpy, "Rotate About Y");
    cube3 = store_cube_instance( dpy, scale_matrix, yrot_matrix, cube, text3, 
	cube_names[2]);
    text4 = store_text( dpy, "Rotate About X and Y");
    cube4 = store_cube_instance( dpy, scale_matrix, rot_matrix, cube, text4, 
	cube_names[3]);

    /* Get root structure and store all 4 instances of the cube in it */
    root = PEXCreateStructure( dpy);

    /* Draw first cube and text in upper-left corner */
    trans_vector.x = -0.25; trans_vector.y = 0.25; trans_vector.z = 0.0;
    PEXTranslate( &trans_vector, trans_matrix);
    PEXSetGlobalTransform( dpy, root, PEXOCStore, trans_matrix);
    PEXExecuteStructure( dpy, root, PEXOCStore, cube1);

    /* Draw second cube and text in upper-right corner */
    trans_vector.x = 0.25; trans_vector.y = 0.25; trans_vector.z = 0.0;
    PEXTranslate( &trans_vector, trans_matrix);
    PEXSetGlobalTransform( dpy, root, PEXOCStore, trans_matrix);
    PEXExecuteStructure( dpy, root, PEXOCStore, cube2);

    /* Draw third cube and text in lower-left corner */
    trans_vector.x = -0.25; trans_vector.y = -0.25; trans_vector.z = 0.0;
    PEXTranslate( &trans_vector, trans_matrix);
    PEXSetGlobalTransform( dpy, root, PEXOCStore, trans_matrix);
    PEXExecuteStructure( dpy, root, PEXOCStore, cube3);

    /* Draw fourth cube and text in lower-right corner */
    trans_vector.x = 0.25; trans_vector.y = -0.25; trans_vector.z = 0.0;
    PEXTranslate( &trans_vector, trans_matrix);
    PEXSetGlobalTransform( dpy, root, PEXOCStore, trans_matrix);
    PEXExecuteStructure( dpy, root, PEXOCStore, cube4);

    return root;
}

int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done, i;   /*indicates when program is done*/
   char				key_pressed;
   PEXStructure			root;
   int                          vis[NUM_CUBES];
   int                          high[NUM_CUBES];
   int                          subset;

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);

   /* Make sure the structure subset is supported */
   subset = ext_info->subset_info & 0xff;
   if ( (subset != 0) && !(subset & PEXStructureMode) ) {
      fprintf( stderr, "Structures are not supported\n");
      exit( 1);
   }

   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
    win = create_window( dpy, TITLE, BlackPixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   /* Ask for key press, mouse press and window exposure events */
   XSelectInput( dpy, win, KeyPressMask | ButtonPressMask | ExposureMask);

   /* Enable z-buffering */
   rd_flags = PEXRAHLHSRMode;
   rd_attr.hlhsr_mode = PEXHLHSRZBuffer;

   /* If PEX version 5.1 or later then let PEX clear the window to 
    * white before rendering */
   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb.red = 0.0;
      rd_attr.background_color.value.rgb.green = 0.0;
      rd_attr.background_color.value.rgb.blue = 0.0;
      rd_flags |= PEXRAClearImage | PEXRABackgroundColor;
   }

   /* Create invisibility and highlight filters */
   rd_attr.highlight_incl = PEXCreateNameSet( dpy);
   rd_attr.invisibility_incl = PEXCreateNameSet( dpy);
   rd_flags |= PEXRAHighlightIncl | PEXRAInvisibilityIncl;

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Create structure store containing cubes */
   root = store_cubes( dpy);

   /* Event loop */
   done = False;
   for (i = 0; i < NUM_CUBES; i++) {
      vis[i] = True;
      high[i] = False;
   }
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window if we are talking to a PEX 5.0 server
          *    Redraw the lines
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
               if ( ext_info->minor_version < 1 )
                  XClearWindow( dpy, win);
               PEXRenderNetwork( dpy, win, rd, root);
            break;

         /*
          * Toggle visibility of cube n whenever 'n' pressed
          *   (where n is either 1,2,3,4).
          * Toggle highlighting of cube whenenver 'a','b','c'or'd' pressed
          *   (a is for cube1, b for cube2, c for cube3 and d for cube4)
          */
         case KeyPress:
            XLookupString((XKeyEvent *)&event, &key_pressed, 1, NULL, NULL);
            if (key_pressed == '1' || key_pressed == '2' ||
                key_pressed == '3' || key_pressed == '4') {
                i = key_pressed - '1';
                if (vis[i] == True) {
                    PEXChangeNameSet( dpy, rd_attr.invisibility_incl,
                        PEXNSAdd, 1, &cube_names[i]);
                    vis[i] = False;
                } else {
                    PEXChangeNameSet( dpy, rd_attr.invisibility_incl,
                        PEXNSRemove, 1, &cube_names[i]);
                    vis[i] = True;
                }
                if ( ext_info->minor_version < 1 ) XClearWindow( dpy, win);
                PEXRenderNetwork( dpy, win, rd, root);
            }
            else if (key_pressed == 'a' || key_pressed == 'b' ||
                     key_pressed == 'c' || key_pressed == 'd') {
                i = key_pressed - 'a';
                if (high[i] == True) {
                    PEXChangeNameSet( dpy, rd_attr.highlight_incl,
                        PEXNSRemove, 1, &cube_names[i]);
                    high[i] = False;
                } else {
                    PEXChangeNameSet( dpy, rd_attr.highlight_incl,
                        PEXNSAdd, 1, &cube_names[i]);
                    high[i] = True;
                }
                if ( ext_info->minor_version < 1 ) XClearWindow( dpy, win);
                PEXRenderNetwork( dpy, win, rd, root);
            }
	    break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   XCloseDisplay( dpy);
   exit( 0);
}
