/*
 *  File Name: cube_pick.c
 *
 *  Description: Picking example from Chapter 14
 * 
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"

#define	TITLE 		"PEX Cube Instances"

#define VERT_PER_FACE   4
#define NUM_FACES	6
static PEXCoord cube_vertices[VERT_PER_FACE*NUM_FACES] = 
{
    { 0.0,  0.0,  1.0}, /* front vertices */
    { 1.0,  0.0,  1.0},
    { 1.0,  1.0,  1.0},
    { 0.0,  1.0,  1.0},
    { 1.0,  0.0,  1.0}, /* right face */
    { 1.0,  0.0,  0.0},
    { 1.0,  1.0,  0.0},
    { 1.0,  1.0,  1.0},
    { 0.0,  0.0,  1.0}, /* left vertices */
    { 0.0,  0.0,  0.0},
    { 0.0,  1.0,  0.0},
    { 0.0,  1.0,  1.0},
    { 0.0,  0.0,  0.0}, /* back face */
    { 1.0,  0.0,  0.0},
    { 1.0,  1.0,  0.0},
    { 0.0,  1.0,  0.0},
    { 0.0,  1.0,  1.0}, /* top face */
    { 1.0,  1.0,  1.0},
    { 1.0,  1.0,  0.0},
    { 0.0,  1.0,  0.0},
    { 0.0,  0.0,  1.0}, /* bottom face */
    { 1.0,  0.0,  1.0},
    { 1.0,  0.0,  0.0},
    { 0.0,  0.0,  0.0}, 
};
static PEXCoord cube_corner = {0.0, 0.0, 0.0};
static PEXCoord cube_center = {0.5, 0.5, 0.5};

PEXColorRGBNormal face_data[NUM_FACES] =
{/* color	                normal */
 {{1.0, 0.0, 0.0},/*red*/     { 0.0, 0.0,  1.0}}, /* front face */
 {{0.0, 1.0, 0.0},/*green*/   { 1.0, 0.0,  0.0}}, /* right face */
 {{0.0, 0.0, 1.0},/*blue*/    {-1.0, 0.0,  0.0}}, /* left face */
 {{1.0, 1.0, 0.0},/*yellow*/  { 0.0, 0.0, -1.0}}, /* back face */
 {{1.0, 0.0, 1.0},/*magenta*/ { 1.0, 0.0,  0.0}}, /* top face */
 {{0.0, 1.0, 1.0},/*cyan*/    {-1.0, 0.0,  0.0}}, /* bottom face */
};

static PEXColorRGB  white_color = {1.0, 1.0, 1.0};
static PEXColorRGB  black_color = {0.0, 0.0, 0.0};
static PEXColorRGB  red_color = {1.0, 0.0, 0.0};

static PEXStructure cube1, cube2, cube3, cube4;

#define NUM_CUBES 4
PEXName cube_names[NUM_CUBES] = {1, 2, 3, 4};

PEXStructure store_cube( dpy)
Display *dpy; 
{
    PEXStructure sid;
    int i,j;
    PEXArrayOfVertex vertices;

    sid = PEXCreateStructure( dpy);

    /* Now store the surface data in the structure */
    for ( i = 0, j = 0; i < NUM_FACES; i++, j+=VERT_PER_FACE) {
        vertices.no_data = &cube_vertices[j];
        PEXFillAreaWithData( dpy, sid, PEXOCStore, PEXShapeConvex, False,
            PEXGAColor | PEXGANormal, PEXGANone, PEXColorTypeRGB,
            (PEXFacetData *)&face_data[i], VERT_PER_FACE, vertices);
    }

    return sid;
}

PEXStructure store_text( dpy, str)
Display *dpy; 
char *str;
{
    PEXStructure sid;
    PEXCoord offset;

    /* offset text from lower left corner of cubes */
    offset.x = offset.y = -.10; offset.z = 0.0;

    sid = PEXCreateStructure( dpy);

    PEXAnnotationText( dpy, sid, PEXOCStore, &cube_corner, &offset,
        strlen(str), str);

    return sid;
}

#define START_TEXT 1
#define START_CUBE 2

PEXStructure store_cube_instance( dpy, scale_matrix, rot_matrix, cube, text, name)
Display *dpy; 
PEXMatrix scale_matrix;
PEXMatrix rot_matrix;
PEXStructure cube; 
PEXStructure text;
PEXName name;
{
    PEXStructure sid;

    sid = PEXCreateStructure( dpy);

    /* Scale data so it fits in view volume */
    PEXSetLocalTransform( dpy, sid, PEXOCStore, PEXReplace, scale_matrix);

    /* Set text attributes (to draw red text) and execute text */
    PEXLabel( dpy, sid, PEXOCStore, START_TEXT);
    PEXSetTextColor( dpy, sid, PEXOCStore, PEXColorTypeRGB, 
        (PEXColor *) &red_color);
    PEXSetATextHeight( dpy, sid, PEXOCStore, .020);
    PEXSetTextPrecision( dpy, sid, PEXOCStore, PEXStrokePrecision);
    PEXExecuteStructure( dpy, sid, PEXOCStore, text);

    /* Set rotation for cube .. must postconcatentate so scale is done first */
    if (rot_matrix != NULL)
       PEXSetLocalTransform( dpy, sid, PEXOCStore, PEXPostConcatenate, rot_matrix);

    /* Set the surface attributes and then execute the cube */
    PEXLabel( dpy, sid, PEXOCStore, START_CUBE);
    PEXSetInteriorStyle( dpy, sid, PEXOCStore, PEXInteriorStyleSolid);
    PEXSetSurfaceEdgeFlag( dpy, sid, PEXOCStore, PEXOn);
    PEXSetSurfaceEdgeColor( dpy, sid, PEXOCStore, PEXColorTypeRGB,
        (PEXColor *) &black_color);
    PEXSetSurfaceEdgeWidth( dpy, sid, PEXOCStore, 2.0); 

    PEXAddToNameSet( dpy, sid, PEXOCStore, 1, &name);
    PEXExecuteStructure( dpy, sid, PEXOCStore, cube);
    PEXRemoveFromNameSet( dpy, sid, PEXOCStore, 1, &name);

    return sid;
}

PEXStructure store_cubes( dpy)
Display *dpy;
{
    PEXMatrix scale_matrix, trans_matrix, rot_matrix, xrot_matrix, yrot_matrix;
    PEXVector trans_vector, scale_vector;
    PEXCoord center;
    PEXStructure text1, text2, text3, text4;
    PEXStructure root, cube;

    /* 
     * Data ranges from (0,0,0) to (1,1,1). It must be scaled 
     * so we can fit four cubes in the viewing volume. We use
     * BuildTransform so we can scale about center of cube
     */
    scale_vector.x = scale_vector.y = scale_vector.z = 0.2;
    trans_vector.x = trans_vector.y = trans_vector.z = 0.0;
    PEXBuildTransform (&cube_center, &trans_vector, 0.0, 0.0, 0.0,
        &scale_vector, scale_matrix);

    /* 
     * Get three different rotation matrices --one to rotate about X,
     * one about Y and one about X and Y. Rotations need to be about
     * cube center.
     */
    scale_vector.x = scale_vector.y = scale_vector.z = 1.0;
    PEXBuildTransform( &cube_center, &trans_vector, M_PI/6.0, 0.0, 0.0,
        &scale_vector, xrot_matrix);
    PEXBuildTransform( &cube_center, &trans_vector, 0.0, M_PI/6.0, 0.0,
        &scale_vector, yrot_matrix);
    PEXBuildTransform( &cube_center, &trans_vector, M_PI/6.0, M_PI/6.0, 0.0,
        &scale_vector, rot_matrix);

    /* Put cube in structure so we can reference it multiple times */
    cube = store_cube( dpy);

    /* Now create cubes with different orientations and label them */
    text1 = store_text( dpy, "No Rotation");
    cube1 = store_cube_instance( dpy, scale_matrix, NULL, cube, text1, 
	cube_names[0]);
    text2 = store_text( dpy, "Rotate About X");
    cube2 = store_cube_instance( dpy, scale_matrix, xrot_matrix, cube, text2, 
	cube_names[1]);
    text3 = store_text( dpy, "Rotate About Y");
    cube3 = store_cube_instance( dpy, scale_matrix, yrot_matrix, cube, text3, 
	cube_names[2]);
    text4 = store_text( dpy, "Rotate About X and Y");
    cube4 = store_cube_instance( dpy, scale_matrix, rot_matrix, cube, text4, 
	cube_names[3]);

    /* Get root structure and store all 4 instances of the cube in it */
    root = PEXCreateStructure( dpy);

    /* Draw first cube and text in upper-left corner */
    trans_vector.x = -0.25; trans_vector.y = 0.25; trans_vector.z = 0.0;
    PEXTranslate( &trans_vector, trans_matrix);
    PEXSetGlobalTransform( dpy, root, PEXOCStore, trans_matrix);
    PEXExecuteStructure( dpy, root, PEXOCStore, cube1);

    /* Draw second cube and text in upper-right corner */
    trans_vector.x = 0.25; trans_vector.y = 0.25; trans_vector.z = 0.0;
    PEXTranslate( &trans_vector, trans_matrix);
    PEXSetGlobalTransform( dpy, root, PEXOCStore, trans_matrix);
    PEXExecuteStructure( dpy, root, PEXOCStore, cube2);

    /* Draw third cube and text in lower-left corner */
    trans_vector.x = -0.25; trans_vector.y = -0.25; trans_vector.z = 0.0;
    PEXTranslate( &trans_vector, trans_matrix);
    PEXSetGlobalTransform( dpy, root, PEXOCStore, trans_matrix);
    PEXExecuteStructure( dpy, root, PEXOCStore, cube3);

    /* Draw fourth cube and text in lower-right corner */
    trans_vector.x = 0.25; trans_vector.y = -0.25; trans_vector.z = 0.0;
    PEXTranslate( &trans_vector, trans_matrix);
    PEXSetGlobalTransform( dpy, root, PEXOCStore, trans_matrix);
    PEXExecuteStructure( dpy, root, PEXOCStore, cube4);

    return root;
}

int set_echo_color( dpy, win, rd, color)
Display *dpy;
Window win;
PEXRenderer rd;
PEXColorRGB *color;
{
   /* Check if echo color escape is supported */
   if ( !check_enum( dpy, win, PEXETEscape, PEXEscapeSetEchoColor) )
      return False;
   PEXSetEchoColor (dpy, rd, PEXColorTypeRGB, (PEXColor *)color);
   return True;
}

static PEXElementRef *path = NULL;
static int npath = 0;

int echo_cube_instance (dpy, win, rd, npick_path, pick_path)
Display *dpy;
Window win;
PEXRenderer rd;
int npick_path;
PEXPickElementRef *pick_path;
{
    PEXRendererAttributes rd_attr;
    int i;

    /* If a cube has been echoed then unecho it */
    if ( path != NULL ) {
        rd_attr.echo_mode = PEXUnEcho;
        PEXChangeRenderer ( dpy, rd, PEXRAEchoMode, &rd_attr);

        /* Now render cube specified by path */
        PEXBeginRendering( dpy, win, rd);
           PEXAccumulateState( dpy, rd, npath, path);
           PEXRenderElements( dpy, rd, path[npath-1].structure, PEXBeginning,
              path[npath-1].offset, PEXBeginning, path[npath-1].offset);
        PEXEndRendering( dpy, rd, True);
       
        free( path);
    }

    /* Convert new pick path to a list of element ref */
    npath = npick_path;
    path = (PEXElementRef *) malloc( sizeof( PEXElementRef) * npath);
    for (i = 0; i < npath; i++) {
       path[i].structure = pick_path[i].sid;
       path[i].offset = pick_path[i].offset;
    }

    /* Now echo cube specified by path */
    rd_attr.echo_mode = PEXEcho;
    PEXChangeRenderer ( dpy, rd, PEXRAEchoMode, &rd_attr);
    PEXBeginRendering( dpy, win, rd);
       PEXAccumulateState( dpy, rd, npath, path);
       PEXRenderElements( dpy, rd, path[npath-1].structure, PEXBeginning,
          path[npath-1].offset, PEXBeginning, path[npath-1].offset);
    PEXEndRendering( dpy, rd, True);
    XSync( dpy, 0);
}

int draw_cubes(dpy, win, rd, root)
Display *dpy;
Window win;
PEXRenderer rd;
PEXStructure root;
{
    PEXRendererAttributes	rd_attr;

    /* Turn on clearing and z-buffering when redrawing entire image */
    /* Make sure echo is turned off */
    rd_attr.clear_image = rd_attr.clear_z = True;
    rd_attr.hlhsr_mode = PEXHLHSRZBuffer;
    rd_attr.echo_mode = PEXNoEcho;
    PEXChangeRenderer( dpy, rd, PEXRAClearImage |  PEXRAClearZ | 
	PEXRAHLHSRMode | PEXRAEchoMode, &rd_attr);

    PEXRenderNetwork( dpy, win, rd, root);

    /* Turn off clearing and z-buffering for incremental drawing */
    rd_attr.clear_image = rd_attr.clear_z = False;
    rd_attr.hlhsr_mode = PEXHLHSROff;
    PEXChangeRenderer ( dpy, rd, PEXRAClearImage |  PEXRAClearZ | 
	PEXRAHLHSRMode, &rd_attr);
}

int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done, i;   /*indicates when program is done*/
   char				key_pressed;
   PEXStructure			root;
   int				x, y;
   int				pick_status, pick_more;
   unsigned long		pick_count;
   PEXPickRecord                pick_record;
   PEXElementRef		start_path;
   PEXPickPath			*pick_path;
   int				subset;

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);

   /* Make sure the structure subset is supported */
   subset = ext_info->subset_info & 0xff;
   if ( (subset != 0) && !(subset & PEXStructureMode) ) {
      fprintf( stderr, "Structures are not supported\n");
      exit( 1);
   }

   if ( ext_info->minor_version == 0 ) {
      fprintf( stderr, "Picking is not supported in PEX 5.0 \n");
      exit( 1);
   }

   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE, BlackPixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   if ( !check_enum( dpy, win, PEXETPickAllMethod, PEXPickAllAll) ||
        !check_enum( dpy, win, PEXETPickDeviceType, PEXPickDeviceDCHitBox) ) {
       fprintf( stderr, "Needed picking capabilities that are not supported \n");
       exit( -1);
   }

   /* Ask for mouse press and window exposure events */
   XSelectInput( dpy, win, KeyPressMask | ButtonPressMask | ExposureMask | 
	StructureNotifyMask );

   /* We know it is PEX version 5.1 or later so set the background */
   /* (clearI is set before drawing) */
   rd_attr.background_color.type = PEXColorTypeRGB;
   rd_attr.background_color.value.rgb.red = 0.0;
   rd_attr.background_color.value.rgb.green = 0.0;
   rd_attr.background_color.value.rgb.blue = 0.0;
   rd_flags = PEXRABackgroundColor;

   /* Make all cubes selectable */
   rd_attr.pick_incl = PEXCreateNameSet( dpy);
   rd_flags |= PEXRAPickIncl;
   PEXChangeNameSet( dpy, rd_attr.pick_incl, PEXNSAdd, NUM_CUBES, cube_names);

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Set the echo color if the escape is supported */
   set_echo_color( dpy, win, rd, &white_color);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Create structure store containing cubes */
   root = store_cubes( dpy);

   /* Event loop */
   done = False;
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Redraw the cubes
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
            draw_cubes( dpy, win, rd, root);
            break;

	 /* 
          * Need to notice when window size changes so we can compute the
          * aperture correctly    
          */
         case ConfigureNotify:
            width = event.xconfigure.width;
            height = event.xconfigure.height;
            break;

	 /* Echo cubes that are picked */
         case ButtonPress:
	    /* 
             * Use device coordinate pick box that is 5% of window   
             * Note that for PEX the origin of device coordinates is 
             * the lower-left corner and for X it is the upper-left 
             */
            pick_record.box.position.x = event.xbutton.x;
            pick_record.box.position.y = height - event.xbutton.y - 1;
            pick_record.box.distance = (float) width/40;

            /* Set pick path */
            start_path.offset = 0;
            start_path.structure = root;
            rd_attr.pick_start_path.count = 1;
            rd_attr.pick_start_path.elements = &start_path;
    	    PEXChangeRenderer ( dpy, rd, PEXRAPickStartPath, &rd_attr);
            
     	    pick_path = PEXPickAll( dpy, win, rd, PEXPickAllAll, 
		1/*max hits*/, PEXPickDeviceDCHitBox, &pick_record, 
                &pick_status, &pick_more, &pick_count);

            if ( pick_status == PEXPick ) {
                echo_cube_instance( dpy, win, rd, pick_path->count, 
		    pick_path->elements);
                PEXFreePickPaths( pick_count, pick_path);
            }
            break;

         /*
          * Typing 'q' or 'Q' exits application.
          */
         case KeyPress:
            XLookupString((XKeyEvent *)&event, &key_pressed, 1, NULL, NULL);
            if (key_pressed == 'q' || key_pressed == 'Q') {
 	 	done = True;
	    }
            break;
      }
   }

   /* Termination */
   if ( path != NULL ) free( path);
   XCloseDisplay( dpy);
   exit( 0);
}
