/*
 *  File Name: cube_views.c
 *
 *  Description: Viewing example from Chapter 10
 * 
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "view_utils.h"
#include "util_protos.h"

#define	TITLE 		"PEX Cube Views"

#define FRONT_VIEW_INDEX 1
#define TOP_VIEW_INDEX 2
#define PERSPECTIVE_VIEW_INDEX 3
#define ISO_VIEW_INDEX 4

#define MINX 0.0
#define MINY 0.0
#define MINZ 0.0
#define MAXX 1.0
#define MAXY 1.0
#define MAXZ 1.0
#define N_CUBE_VERT 8
static PEXCoord cube_vertices[N_CUBE_VERT] =
{
    {MINX, MINY, MAXZ}, /* front vertices */
    {MAXX, MINY, MAXZ},
    {MAXX, MAXY, MAXZ},
    {MINX, MAXY, MAXZ},
    {MINX, MINY, MINZ}, /* back vertices */
    {MAXX, MINY, MINZ},
    {MAXX, MAXY, MINZ},
    {MINX, MAXY, MINZ},
};

#define VERT_PER_FACE   4
#define NUM_FACES	6
unsigned short indices[VERT_PER_FACE * NUM_FACES] =
{
    0, 1, 2, 3,     /* front face */
    1, 5, 6, 2,     /* right face */
    0, 4, 7, 3,     /* left face */
    4, 5, 6, 7,     /* back face */
    3, 2, 6, 7,     /* top face */
    0, 1, 5, 4,     /* bottom face */
};

PEXColorRGBNormal face_data[NUM_FACES] =
{
 /* color                       normal  */
 {{1.0, 0.0, 0.0},/*red*/     { 0.0, 0.0,  1.0}}, /* front face */
 {{0.0, 1.0, 0.0},/*green*/   { 1.0, 0.0,  0.0}}, /* right face */
 {{0.0, 0.0, 1.0},/*blue*/    {-1.0, 0.0,  0.0}}, /* left face */
 {{1.0, 1.0, 0.0},/*yellow*/  { 0.0, 0.0, -1.0}}, /* back face */
 {{1.0, 0.0, 1.0},/*magenta*/ { 1.0, 0.0,  0.0}}, /* top face */
 {{0.0, 1.0, 1.0},/*cyan*/    {-1.0, 0.0,  0.0}}, /* bottom face */
};

#define N_BORDER_VERT 5
static PEXCoord2D border[N_BORDER_VERT] =
{
   {0.02,  0.02},
   {0.02,  0.98},
   {0.98,  0.98},
   {0.98,  0.02},
   {0.02,  0.02},
};
static PEXCoord2D viewport_corner = {0.02, 0.02};

static PEXColorRGB  red_color = {1.0, 0.0, 0.0};
static PEXColorRGB  black_color = {0.0, 0.0, 0.0};

PEXStructure store_cube( dpy)
Display *dpy; 
{
    PEXStructure sid;
    int i;
    PEXListOfUShort fill_areas[NUM_FACES];
    PEXConnectivityData fill_area_sets[NUM_FACES];
    unsigned short *fill_area_indices = indices;
    PEXArrayOfFacetData facets;
    PEXArrayOfVertex    vertices;

    /*
     * Initialize the fill areas--there is one for each face.
     * Also initialize the fill area sets. Each fill area
     * set just contains one fill area (i.e. one face )
     */
    for ( i = 0; i < NUM_FACES; i++) {
      fill_areas[i].count = VERT_PER_FACE;
      fill_areas[i].shorts = fill_area_indices;
      fill_area_indices += VERT_PER_FACE;

      fill_area_sets[i].count = 1; /* just one fill area per set */
      fill_area_sets[i].lists = &fill_areas[i];
    }
    facets.rgb_normal = face_data;
    vertices.no_data = cube_vertices;

    sid = PEXCreateStructure( dpy);

    /* Now store the surface data in the structure */
    PEXSetOfFillAreaSets( dpy, sid, PEXOCStore, PEXShapeConvex, 
        PEXGAColor | PEXGANormal, PEXGANone, PEXGANone, PEXContourDisjoint, 
        True/*one fill area per contour*/, PEXColorTypeRGB,
        NUM_FACES, facets, N_CUBE_VERT, vertices, 
        VERT_PER_FACE * NUM_FACES, NULL/*no edge data*/, fill_area_sets);

    return sid;
}

void draw_wireframe_cube( dpy, rd, cube, view_index)
Display *dpy; 
PEXRenderer rd;
PEXStructure cube;
unsigned int view_index;
{
    PEXSetViewIndex( dpy, rd, PEXOCRender, view_index);

    /* Draw hollow cube */
    PEXSetInteriorStyle( dpy, rd, PEXOCRender, PEXInteriorStyleHollow);
    PEXExecuteStructure( dpy, rd, PEXOCRender, cube);
}

void draw_solid_cube( dpy, rd, cube, view_index)
Display *dpy; 
PEXRenderer rd;
PEXStructure cube;
unsigned int view_index;
{
    PEXSetViewIndex( dpy, rd, PEXOCRender, view_index);

    /* Draw solid cube with black edges */
    PEXSetInteriorStyle( dpy, rd, PEXOCRender, PEXInteriorStyleSolid);
    PEXSetSurfaceEdgeFlag( dpy, rd, PEXOCRender, PEXOn);
    PEXSetSurfaceEdgeColor( dpy, rd, PEXOCRender, PEXColorTypeRGB,
        (PEXColor *)&black_color);
    PEXSetSurfaceEdgeWidth( dpy, rd, PEXOCRender, 2.0);

    PEXExecuteStructure( dpy, rd, PEXOCRender, cube);
}

void draw_viewport( dpy, rd, str)
Display *dpy;
PEXRenderer rd;
char *str;
{
    PEXCoord2D offset;

    /* offset of viewport label from lower-left corner */
    offset.x = offset.y = .05;

    PEXSetLineType( dpy, rd, PEXOCRender, PEXLineTypeDashed);
    PEXSetLineColor(dpy, rd, PEXOCRender, PEXColorTypeRGB, 
        (PEXColor *)&red_color);
    PEXPolyline2D( dpy, rd, PEXOCRender, N_BORDER_VERT, border);

    PEXSetTextColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
        (PEXColor *)&red_color);
    PEXSetATextHeight( dpy, rd, PEXOCRender, .04);
    PEXSetTextPrecision( dpy, rd, PEXOCRender, PEXStrokePrecision);
    PEXAnnotationText2D( dpy, rd, PEXOCRender, &viewport_corner, 
        &offset, strlen(str), str);
}

PEXStructure draw_views( dpy, win, rd, width, height, cube)
Display *dpy;
Window win;
PEXRenderer rd;
int width;
int height;
PEXStructure cube;
{
    int mid_x, mid_y;
    PEXRendererAttributes rd_attr;

    /* compute middle of drawable */
    mid_x = width / 2;
    mid_y = height / 2;

    rd_attr.viewport.use_drawable = False;
    rd_attr.viewport.min.z = 0.0;
    rd_attr.viewport.max.z = 1.0;

    /* First view is a front view.. viewport is in upper left corner */
    rd_attr.viewport.min.x = 0;
    rd_attr.viewport.min.y = mid_y;
    rd_attr.viewport.max.x = mid_x - 1;
    rd_attr.viewport.max.y = 2 * mid_y - 1;
    PEXChangeRenderer ( dpy, rd, PEXRAViewport, &rd_attr);
    PEXBeginRendering( dpy, win, rd);
         draw_viewport( dpy, rd, "Front View");
         draw_solid_cube( dpy, rd, cube, FRONT_VIEW_INDEX);
    PEXEndRendering( dpy, rd, True);

    /* Second view is top view.. viewport is in upper right corner */
    rd_attr.viewport.min.x = mid_x;
    rd_attr.viewport.min.y = mid_y;
    rd_attr.viewport.max.x = 2 * mid_x - 1;
    rd_attr.viewport.max.y = 2 * mid_y - 1;
    PEXChangeRenderer ( dpy, rd, PEXRAViewport, &rd_attr);
    PEXBeginRendering( dpy, win, rd);
         draw_viewport( dpy, rd, "Top View");
         draw_solid_cube( dpy, rd, cube, TOP_VIEW_INDEX);
    PEXEndRendering( dpy, rd, True);

    /* Third view is isometric view.. viewport is in lower left corner */
    rd_attr.viewport.min.x = 0;
    rd_attr.viewport.min.y = 0;
    rd_attr.viewport.max.x = mid_x - 1;
    rd_attr.viewport.max.y = mid_y - 1;
    PEXChangeRenderer ( dpy, rd, PEXRAViewport, &rd_attr);
    PEXBeginRendering( dpy, win, rd);
         draw_viewport( dpy, rd, "Iso View");
         draw_solid_cube( dpy, rd, cube, ISO_VIEW_INDEX);
    PEXEndRendering( dpy, rd, True);

    /* Fourth view is perspective view.. viewport is in lower right corner */
    rd_attr.viewport.min.x = mid_x;
    rd_attr.viewport.min.y = 0;
    rd_attr.viewport.max.x = 2 * mid_x - 1;
    rd_attr.viewport.max.y = mid_y - 1;
    PEXChangeRenderer ( dpy, rd, PEXRAViewport, &rd_attr);
    PEXBeginRendering( dpy, win, rd);
         draw_viewport( dpy, rd, "Perspective View");
         draw_wireframe_cube( dpy, rd, cube, PERSPECTIVE_VIEW_INDEX);
    PEXEndRendering( dpy, rd, True);
}

int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   int				use_defaults;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done;   /*indicates when program is done*/
   PEXStructure 	        cube;
   int				subset;

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);

   /* Make sure the structure subset is supported */
   subset = ext_info->subset_info & 0xff;
   if ( (subset != 0) && !(subset & PEXStructureMode) ) {
      fprintf( stderr, "Structures are not supported\n");
      exit( 1);
   }

   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE, WhitePixel( dpy, DefaultScreen( dpy)),
          &vis_info, &cmap_info, &width, &height);


   /* Ask for mouse press and window exposure events */
   XSelectInput( dpy, win, ButtonPressMask | StructureNotifyMask | ExposureMask);

   /* Create multipe views all with object in the center */
   rd_attr.view_table = PEXCreateLookupTable( dpy, win, PEXLUTView);
   rd_flags = PEXRAViewTable;

   init_object_view( dpy, MINX, MAXX, MINY, MAXY, MINZ, MAXZ,
      .75, ORTHOGRAPHIC_PROJECTION, FRONT_VIEW, FRONT_VIEW_INDEX,
      rd_attr.view_table);
   init_object_view( dpy, MINX, MAXX, MINY, MAXY, MINZ, MAXZ,
      .75, ORTHOGRAPHIC_PROJECTION, TOP_VIEW, TOP_VIEW_INDEX,
      rd_attr.view_table);
   init_object_view( dpy, MINX, MAXX, MINY, MAXY, MINZ, MAXZ,
      .5, ORTHOGRAPHIC_PROJECTION, ISO_VIEW, ISO_VIEW_INDEX,
      rd_attr.view_table);
   init_object_view( dpy, MINX, MAXX, MINY, MAXY, MINZ, MAXZ,
      .75, PERSPECTIVE_PROJECTION, LEFT_SIDE_VIEW, PERSPECTIVE_VIEW_INDEX,
      rd_attr.view_table);

   /* Enable z-buffering */
   rd_flags |= PEXRAHLHSRMode;
   rd_attr.hlhsr_mode = PEXHLHSRZBuffer;

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Create structure store containing cubes */
   cube = store_cube( dpy);

   /* Event loop */
   done = False;
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window 
          *    Redraw the cubes
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
            XClearWindow( dpy, win);
            draw_views( dpy, win, rd, width, height, cube);
            break;

         case ConfigureNotify: 
            if (event.xconfigure.width < event.xconfigure.height )
                width = height = event.xconfigure.width;
            else
                width = height = event.xconfigure.height;
            break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   XCloseDisplay( dpy);
   exit( 0);
}
