/*
 *  File Name: fillareas.c
 *
 *  Description: Fill areas example from Chapter 7.
 * 
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"

#define	TITLE 		"PEX Fill Areas"

/* 
 * Used to draw fill area
 */
static PEXCoord triangle[3] =
{
 {0.2,   0.7,   0.0},
 {0.8,   0.7,   0.0},
 {0.5,   0.9,   0.0},
};

/* 
 * Used to draw fill area set
 */
static PEXCoord outer_quad[4] =
{
 {0.2,   0.2,   0.0},
 {0.8,   0.2,   0.0},
 {0.8,   0.5,   0.0},
 {0.2,   0.5,   0.0},
};
static PEXCoord inner_quad[4] =
{
 {0.4,   0.3,   0.0},
 {0.6,   0.3,   0.0},
 {0.6,   0.4,   0.0},
 {0.4,   0.4,   0.0},
};

static PEXColorRGB  red_color = {1.0, 0.0, 0.0};
static PEXColorRGB  black_color = {0.0, 0.0, 0.0};

/* 
 * Fill area example in chapter 7
 */
int draw_fill_areas( dpy, win, rd )
Display *dpy; 
Window win; 
PEXRenderer rd;
{
    PEXCoord offset, position;
    PEXListOfCoord   fas[2];

    PEXBeginRendering( dpy, win, rd);
        /* Draw hollow fill area */
	PEXSetSurfaceColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
	    (PEXColor *)&red_color);
	PEXSetInteriorStyle( dpy, rd, PEXOCRender, PEXInteriorStyleHollow);
        PEXFillArea( dpy, rd, PEXOCRender, PEXShapeConvex, False, 3, triangle); 

        /* Label Fill Area */
	PEXSetTextColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
	    (PEXColor *)&black_color);
        PEXSetTextPrecision( dpy, rd, PEXOCRender, PEXStrokePrecision);
        PEXSetATextHeight( dpy, rd, PEXOCRender, .02);
 	position = triangle[0];
        offset.x = offset.z = 0.0; offset.y = -.1;
        PEXAnnotationText( dpy, rd, PEXOCRender, &position, &offset, 
		 strlen("Hollow Fill Area"), "Hollow Fill Area");

        /* Draw solid fill area set with edges on */
	PEXSetSurfaceEdgeColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
	    (PEXColor *)&black_color);
	PEXSetSurfaceEdgeWidth( dpy, rd, PEXOCRender, 2.0);
	PEXSetInteriorStyle( dpy, rd, PEXOCRender, PEXInteriorStyleSolid);
	PEXSetSurfaceEdgeFlag( dpy, rd, PEXOCRender, PEXOn);
        fas[0].count = 4;
        fas[0].points = outer_quad;
        fas[1].count = 4;
        fas[1].points = inner_quad;
        PEXFillAreaSet( dpy, rd, PEXOCRender, PEXShapeConvex, False, 
		PEXContourNested, 2, fas); 

        /* Label Fill Area Set - using previously set attributes */
 	position = outer_quad[0];
        offset.x = offset.z = 0.0; offset.y = -.1;
        PEXAnnotationText( dpy, rd, PEXOCRender, &position, &offset, 
		 strlen("Solid Fill Area Set With Edges"), 
      		 "Solid Fill Area Set With Edges");
    PEXEndRendering( dpy, rd, True);
}

int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done;   /*indicates when program is done*/

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);
 
   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE, WhitePixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   /* Ask for mouse press and window exposure events */
   XSelectInput( dpy, win, ButtonPressMask | ExposureMask);

   /* If PEX version 5.1 or later then let PEX clear the window to 
    * white before rendering */
   rd_flags = 0;
   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb.red = 1.0;
      rd_attr.background_color.value.rgb.green = 1.0;
      rd_attr.background_color.value.rgb.blue = 1.0;
      rd_flags |= PEXRAClearImage | PEXRABackgroundColor;
   }

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Event loop */
   done = False;
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window if we are talking to a PEX 5.0 server
          *    Redraw the fill areas
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
            if ( ext_info->minor_version < 1 )
               XClearWindow( dpy, win);
            draw_fill_areas( dpy, win, rd);
            break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   XCloseDisplay( dpy);
   exit( 0);
}
