/*
 *  File Name: makeframes.c
 *
 *  Author: David Yu 
 *          Converted to MIT PEXlib by Paula Womack
 *
 *  Revision: 1.0
 */
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/PEX5/PEXlib.h>


#define  PI		(M_PI)
#define  MAX(a, b)	(((a) > (b)) ? (a) : (b))
#define  MIN(a, b)	(((a) < (b)) ? (a) : (b))

PEXVector* normalize(a)
    PEXVector  *a;
{
    float  length;
    float  invLength;

    length = (float) sqrt((double)(a->x * a->x + a->y * a->y + a->z * a->z));
    invLength = 1.0 / length;

    a->x *= invLength;
    a->y *= invLength;
    a->z *= invLength;

    return a;
}


PEXVector* computeTriNormal(a, b, c, norm)
    PEXCoord  *a;
    PEXCoord  *b;
    PEXCoord  *c;
    PEXVector *norm;
{
    norm->x = (b->y - a->y) * (c->z - a->z) - (c->y - a->y) * (b->z - a->z);
    norm->y = (b->z - a->z) * (c->x - a->x) - (c->z - a->z) * (b->x - a->x);
    norm->z = (b->x - a->x) * (c->y - a->y) - (c->x - a->x) * (b->y - a->y);

    return normalize(norm);
}


evalSurface(frame, u, v, numFrames, minAmp, maxAmp, wavelength, pPoint, pNorm)
    int         frame;
    float       u;
    float       v;
    float       minAmp;
    float       maxAmp;
    float       wavelength;
    PEXCoord    *pPoint;
    PEXVector   *pNorm;
{
    float  distance;
    float  amplitude;
    float  angle;
    float  phaseShift;
    float  slope;
    float  sign;

    distance = (float) sqrt((double) (u*u + v*v));
    amplitude = maxAmp - distance * (maxAmp - minAmp);
    angle = distance * (2.0 * PI / wavelength);
    phaseShift = frame * (2.0 * PI / numFrames);

    /*
     * compute point on surface
     */
    pPoint->x = u;
    pPoint->y = v;
    pPoint->z = amplitude * (float) cos((double) (angle + phaseShift));

    /*
     * compute slope of surface
     */
    slope = amplitude * (float) -sin((double) (angle + phaseShift));

    /*
     * compute normal at point on surface
     */
#define NEAR_ZERO (.00001)
    if ( (slope > NEAR_ZERO || slope < -NEAR_ZERO) && distance > NEAR_ZERO ) 
    {
	sign = (slope < 0.0) ? -1.0 : 1.0;
	pNorm->x = sign * (-u / (distance * maxAmp));
	pNorm->y = sign * (-v / (distance * maxAmp));
	pNorm->z = sign / slope;
	normalize(pNorm);
    }
    else
    {
	pNorm->x = 0.0;
	pNorm->y = 0.0;
	pNorm->z = 1.0;
    }
}


sendTriStrip(dpy, str, uRes, vRes, facetColors, vertPoints, vertNorms)
    Display            *dpy;
    PEXStructure       str;
    int                uRes;
    int                vRes;
    PEXColorRGB	       *facetColors;
    PEXCoord           *vertPoints;
    PEXVector          *vertNorms;
{
    PEXColorRGBNormal 	*facetData, *pFacet;
    PEXVertexNormal 	*vertData, *pVert;
    int             	i, j;
    int             	loIndex;
    int             	hiIndex;
    int             	facetIndex;
    int             	a, b, c, f;
    PEXArrayOfFacetData facets;
    PEXArrayOfVertex    vertices;


    /*
     * allocate storage
     */
    facetData = (PEXColorRGBNormal *) 
	calloc((unsigned) 2 * uRes, sizeof(PEXColorRGBNormal));
    vertData = (PEXVertexNormal *) 
	calloc((unsigned) 2 * uRes + 2, sizeof(PEXVertexNormal));

    for (i = 0; i < vRes; ++i)
    {
	pFacet = facetData;
	pVert = vertData;

	loIndex = i * (uRes + 1);
	hiIndex = loIndex + (uRes + 1);
	facetIndex = i * uRes - 1;

	a = loIndex;
	pVert->point = vertPoints[a];
	pVert->normal =  vertNorms[a];
        pVert++;

	b = hiIndex;
	pVert->point = vertPoints[b];
	pVert->normal = vertNorms[b];
        pVert++;

	for (j = 1; j <= uRes; ++j)
	{
	    f = facetIndex + j;

	    c = loIndex + j;
	    pVert->point = vertPoints[c];
	    pVert->normal = vertNorms[c];
            pVert++;

	    pFacet->rgb = facetColors[f];
	    computeTriNormal(&vertPoints[c], &vertPoints[b], &vertPoints[a],
			     &pFacet->normal); 
            pFacet++;

	    a = b;
	    b = c;

	    c = hiIndex + j;
	    pVert->point = vertPoints[c];
	    pVert->normal = vertNorms[c];
            pVert++;

	    pFacet->rgb = facetColors[f];
	    computeTriNormal(&vertPoints[a], &vertPoints[b], &vertPoints[c],
			     &pFacet->normal);
            pFacet++;

	    a = b;
	    b = c;
	}

	/*
	 * send each row as a triangle strip
	 */
        facets.rgb_normal = facetData;
        vertices.normal = vertData;
	PEXTriangleStrip(dpy, str, PEXOCStore, (PEXGANormal | PEXGAColor),
	     PEXGANormal, PEXColorTypeRGB, facets, 2 * uRes + 2, vertices);
    }

    free(facetData);
    free(vertData);
}


buildwaves(dpy, str, nFrames, wavelen, minAmp, maxAmp, uRes, vRes, c1, c2)
    Display           *dpy;
    PEXStructure      *str;
    int                nFrames;
    float              wavelen;
    float              minAmp;
    float              maxAmp;
    int                uRes;
    int                vRes;
    PEXColorRGB        *c1;
    PEXColorRGB	       *c2;
{
    int                nFacets;
    int                nVerts;
    PEXColorRGB	       *facetColors;
    PEXCoord           *vertPoints;
    PEXVector          *vertNorms;
    int                majRes;
    float              u, u0, du;
    float              v, v0, dv;
    int                i, j;
    int                index;
    int                frame;

    nFacets = uRes * vRes;
    nVerts = (uRes + 1) * (vRes + 1);

    facetColors = (PEXColorRGB *)
	          calloc((unsigned) nFacets, sizeof(PEXColorRGB));
    vertPoints = (PEXCoord *) calloc((unsigned) nVerts, sizeof(PEXCoord));
    vertNorms = (PEXVector *) calloc((unsigned) nVerts, sizeof(PEXVector));

    /*
     * assign checkerboard colors
     */
    for (i = 0; i <= vRes; ++i)
    {
	for (j = 0; j <= uRes; ++j)
	{
	    int  even = ((i / 2) % 2) ^ ((j / 2) % 2);

	    if (i < vRes && j < uRes)
	    {
		index = i * uRes + j;
	        facetColors[index] = (even) ? *c1 : *c2;
	    }
	}
    }

    /*
     * parameterize surface
     */
    majRes = MAX(uRes, vRes);
    du = 1.0 / majRes;
    dv = 1.0 / majRes;
    u0 = (majRes - uRes) / 2 * du;
    v0 = (majRes - vRes) / 2 * dv;

    /*
     * compute wave frames
     */
    for (frame = 0; frame < nFrames; ++frame)
    {
	v = v0;
	for (i = 0; i <= vRes; ++i)
	{
	    u = u0;
	    for (j = 0; j <= uRes; ++j)
	    {
		index = i * (uRes + 1) + j;

		evalSurface(frame, u, v,
			    nFrames, minAmp, maxAmp, wavelen,
			    &vertPoints[index], &vertNorms[index]);

		u += du;
	    }
	    v += dv;
	}

	sendTriStrip(dpy, str[frame], uRes, vRes,
		     facetColors, vertPoints, vertNorms);

    }

    free(facetColors);
    free(vertPoints);
    free(vertNorms);
}

