/*
 *  File Name: pexwave.c
 *
 *  Description: Wave program in Appendix D
 *
 *  Author: David Yu 
 *          Converted to MIT PEXlib and book utilities by Paula Womack
 *
 *  Revision: 1.0
 */

#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"
#include "view_utils.h"


/* macros */
#define PI		M_PI
#define TWO_PI		(2.0 * (PI))
#define DEGRAD(d)	((d) * (PI) / 180.0)


/* defaults */
#define NUM_FRAMES	10
#define REFINEMENT	16
#define X_INITIAL	DEGRAD(-60.0)
#define Y_INITIAL	DEGRAD(0.0)
#define WAVE_LENGTH     0.6
#define MIN_AMPLITUDE   0.1
#define MAX_AMPLITUDE   0.1
#define LABEL		"PEXWAVE"


/* program stuff */
char	       *displayName = NULL;
int		reverse = 0;
int		h_res = REFINEMENT, v_res = REFINEMENT;
unsigned int    winWidth, winHeight;
double          wavelength = WAVE_LENGTH;
double          minAmp = MIN_AMPLITUDE;
double          maxAmp = MAX_AMPLITUDE;
unsigned int	demoMode = 0;
double		rotA = X_INITIAL, rotB = Y_INITIAL;
char		keyPressed;
int		i, frame, lastx, lasty, draw_buf;

/* X stuff */
Display		       *dpy;
Window			win;
XEvent			event;
XVisualInfo             vis_info; /* color information */
XStandardColormap       cmap_info;
XSetWindowAttributes    win_attr;

/* PEX stuff */
PEXExtensionInfo        *ext_info;/* info about PEX extension */
PEXColorApproxEntry     approx_entry;
PEXLookupTable		view_lut;
PEXLookupTable		light_lut;
PEXViewEntry		view;
PEXLightEntry		light;
PEXPipelineContext	pc;
PEXPCAttributes		pc_attr;
unsigned long		masks[3], wave_hlhsr = PEXHLHSROff;
PEXRenderer		rd;
PEXRendererAttributes	rd_attr;
unsigned long           rd_flags;
PEXMatrix		tmpXform;
PEXColorRGB		wave_color1, wave_color2;
PEXStructure		wavestr[NUM_FRAMES];
PEXVector 	        trans, scale;
PEXCoord 		from;
PEXCoord 		to;
PEXVector 		up;
PEXTableIndex	        light_indices[5];
int			num_lights = 0;
bufData                 db_data;
PEXColorRGB  		background_color = {0.0, 0.0, 0.0};
int			subset;

main(argc, argv)
    int     argc;
    char  **argv;
{
    /* parse command line arguments */
    if (argc > 1) {
	for (i = 1; i < argc; ++i) {
	    if ((!strcmp(argv[i], "-d")) && ((i+1) < argc)) {
		displayName = argv[++i];
	    } else if ((!strcmp(argv[i], "-s")) && ((i+2) < argc)) {
		h_res = atoi(argv[++i]);
		v_res = atoi(argv[++i]);
	    } else if ((!strcmp(argv[i], "-demo")) && ((i+1) < argc)) {
		demoMode = atoi(argv[++i]) + 1;
	    } else if (!strcmp(argv[i], "-r")) {
		reverse = !reverse;
	    } else if (!strcmp(argv[i], "-z")) {
		wave_hlhsr = PEXHLHSRZBuffer;
	    } else {
		usage(*argv);
		exit(1);
	    }
	}
    }

    /* Open a connection to the PEX server */
    dpy = open_pex_connection( &ext_info);
    if ( dpy == NULL ) exit( 1);

    /* Make sure the structure subset is supported */
    subset = ext_info->subset_info & 0xff;
    if ( (subset != 0) && !(subset & PEXStructureMode) ) {
      fprintf( stderr, "Structures are not supported\n");
      exit( 1);
    }

    /* Find best visual and colormap. If we fail, just exit */
    if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

    /* Create a window */
    win = create_window( dpy, LABEL, BlackPixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &winWidth, &winHeight);

    /* set event mask */
    if (demoMode) {
	XSelectInput(dpy, win, ExposureMask | KeyPressMask | StructureNotifyMask);
    } else {
	XSelectInput(dpy, win, ExposureMask | KeyPressMask | StructureNotifyMask |
			       ButtonPressMask | ButtonMotionMask);
    }

    /* 
     * Set up a perspective view 
     */
    view_lut = PEXCreateLookupTable(dpy, win, PEXLUTView);
    PEXPerspProjMatrix(DEGRAD(45.0), 9.0, 1.0, 8.0, -8.0, view.mapping);
    from.x = from.y = 0.0; from.z = 9.0;
    to.x = to.y = to.z = 0.0;
    up.x = up.z = 0.0; up.y = 1.0;
    PEXLookAtViewMatrix(&from, &to, &up, view.orientation);
    view.clip_flags = 0;
    view.clip_limits.min.x = view.clip_limits.min.y =
        view.clip_limits.min.z = 0.0;
    view.clip_limits.max.x = view.clip_limits.max.y =
         view.clip_limits.max.z = 1.0;
    PEXSetTableEntries(dpy, view_lut, 1, 1, PEXLUTView, &view);

    /* 
     * Set up one ambient light and 2 vector lights
     */
    light_lut = PEXCreateLookupTable(dpy, win, PEXLUTLight);
    light.type = PEXLightAmbient;
    light.color.type = PEXColorTypeRGB;
    light.color.value.rgb.red = 0.4;
    light.color.value.rgb.green = 0.4;
    light.color.value.rgb.blue = 0.4;
    PEXSetTableEntries(dpy, light_lut, 1, 1, PEXLUTLight, &light);
    light_indices[num_lights++] = 1;

    light.type = PEXLightWCVector;
    light.direction.x = -0.577;
    light.direction.y = -0.577;
    light.direction.z = 0.577;
    light.color.type = PEXColorTypeRGB;
    light.color.value.rgb.red = 1.0;
    light.color.value.rgb.green = 1.0;
    light.color.value.rgb.blue = 1.0;
    PEXSetTableEntries(dpy, light_lut, 2, 1, PEXLUTLight, &light);
    light_indices[num_lights++] = 2;

    light.type = PEXLightWCVector;
    light.direction.x = 0.577;
    light.direction.y = 0.577;
    light.direction.z = -0.577;
    light.color.type = PEXColorTypeRGB;
    light.color.value.rgb.red = 1.0;
    light.color.value.rgb.green = 1.0;
    light.color.value.rgb.blue = 1.0;
    PEXSetTableEntries(dpy, light_lut, 3, 1, PEXLUTLight, &light);
    light_indices[num_lights++] = 3;

    /* 
     * Create a pipeline context to store all the initial state 
     */
    masks[0] = masks[1] = masks[2] = 0;

    trans.x = trans.y = -0.5; trans.z = 0.0;
    PEXTranslate(&trans, pc_attr.local_transform);
    scale.x = scale.y = scale.z = 5.0;
    PEXScale(&scale, tmpXform);
    PEXMatrixMult(tmpXform, pc_attr.local_transform,
			pc_attr.local_transform);
    PEXSetPCAttributeMask(masks, PEXPCLocalTransform);

    PEXRotate(PEXXAxis, X_INITIAL, pc_attr.global_transform);
    PEXRotate(PEXYAxis, Y_INITIAL, tmpXform);
    PEXMatrixMult(tmpXform, pc_attr.global_transform,
			pc_attr.global_transform);
    PEXSetPCAttributeMask(masks, PEXPCGlobalTransform);

    pc_attr.surface_color.type = PEXColorTypeRGB;
    pc_attr.surface_color.value.rgb.red = 0.0;
    pc_attr.surface_color.value.rgb.green = 0.5;
    pc_attr.surface_color.value.rgb.blue = 1.0;
    PEXSetPCAttributeMask(masks, PEXPCSurfaceColor);

    pc_attr.bf_surface_color.type = PEXColorTypeRGB;
    pc_attr.bf_surface_color.value.rgb.red = 0.0;
    pc_attr.bf_surface_color.value.rgb.green = 0.5;
    pc_attr.bf_surface_color.value.rgb.blue = 1.0;
    PEXSetPCAttributeMask(masks, PEXPCBFSurfaceColor);

    pc_attr.view_index = 1;
    PEXSetPCAttributeMask(masks, PEXPCViewIndex);
    pc_attr.culling_mode = PEXNone;
    PEXSetPCAttributeMask(masks, PEXPCCullingMode);
    pc_attr.interior_style = PEXInteriorStyleSolid;
    PEXSetPCAttributeMask(masks, PEXPCInteriorStyle);
    pc_attr.surface_interp = PEXSurfaceInterpColor;
    PEXSetPCAttributeMask(masks, PEXPCSurfaceInterp);

    pc_attr.reflection_model = PEXReflectionSpecular;
    PEXSetPCAttributeMask(masks, PEXPCReflectionModel);
    pc_attr.reflection_attr.ambient = 0.4;
    pc_attr.reflection_attr.diffuse = 0.7;
    pc_attr.reflection_attr.specular = 0.9;
    pc_attr.reflection_attr.specular_conc = 10.0;
    pc_attr.reflection_attr.transmission = 0.0;
    pc_attr.reflection_attr.specular_color.type = PEXColorTypeRGB;
    pc_attr.reflection_attr.specular_color.value.rgb.red = 1.0;
    pc_attr.reflection_attr.specular_color.value.rgb.green = 1.0;
    pc_attr.reflection_attr.specular_color.value.rgb.blue = 1.0;
    PEXSetPCAttributeMask(masks, PEXPCReflectionAttr);

    pc_attr.light_state.count = num_lights;
    pc_attr.light_state.indices = light_indices;
    PEXSetPCAttributeMask(masks, PEXPCLightState);

    pc = PEXCreatePipelineContext(dpy, masks, &pc_attr);

    /* 
     * Create a renderer 
     */
    rd_attr.pipeline_context = pc;
    rd_attr.view_table = view_lut;
    rd_attr.light_table = light_lut;
    rd_attr.hlhsr_mode = wave_hlhsr;
    rd_flags = PEXRAPipelineContext | PEXRAViewTable | PEXRALightTable | 
	      PEXRAHLHSRMode;
    rd = create_renderer(dpy, win, &approx_entry, rd_flags, &rd_attr);

    /* Create back buffers for double buffering */
    draw_buf = create_back_buffer(dpy, win, rd, &db_data, &background_color,
        BlackPixel( dpy, DefaultScreen( dpy)));

    /* Map the window */
    XMapWindow(dpy, win);

    /* 
     * Create a structure for each frame and fill it with wave data 
     */
    /* blue & gray wave */
    wave_color1.red = 0.7;
    wave_color1.green = 0.7;
    wave_color1.blue = 0.7;
    wave_color2.red = 0.0;
    wave_color2.green = 0.0;
    wave_color2.blue = 1.0;

    for (i = 0; i < NUM_FRAMES; ++i)
	wavestr[i] = PEXCreateStructure(dpy);
    buildwaves(dpy, wavestr, NUM_FRAMES, wavelength, minAmp, maxAmp, h_res, 
	v_res, &wave_color1, &wave_color2);

    /* main event loop */
    i = NUM_FRAMES -1;
    while(((demoMode > 0) ? --demoMode : 1)) {

	if (XPending(dpy)) {
	    XNextEvent(dpy, &event);

	    if (event.type == KeyPress) {
		XLookupString( (XKeyEvent *)&event, &keyPressed, 1, NULL, NULL);
		if (keyPressed == '') {
		   exit(0);
		}
	    }

	    if (event.type == ConfigureNotify) {
                winWidth = event.xconfigure.width;
                winHeight = event.xconfigure.height;
	        resize_back_buffer( dpy, winWidth, winHeight, &db_data);
		continue;
	    }

	    if (event.type == ButtonPress) {
		lastx = event.xbutton.x;
		lasty = event.xbutton.y;
		continue;
	    }

	    if (event.type == MotionNotify) {
		rotA = (TWO_PI * (event.xmotion.y - lasty)) / winHeight;
		rotB = (TWO_PI * (event.xmotion.x - lastx)) / winWidth;
		lastx = event.xmotion.x;
		lasty = event.xmotion.y;
		PEXRotate(PEXXAxis, rotA, tmpXform);
		PEXMatrixMult(tmpXform, pc_attr.global_transform,
				    pc_attr.global_transform);
		PEXRotate(PEXYAxis, rotB, tmpXform);
		PEXMatrixMult(tmpXform, pc_attr.global_transform,
				    pc_attr.global_transform);
		masks[0] = masks[1] = masks[2] = 0;
		PEXSetPCAttributeMask(masks, PEXPCGlobalTransform);
		PEXChangePipelineContext(dpy, pc, masks, &pc_attr);
		continue;
	    }
	}

	if (demoMode) {
	    PEXRotate(PEXZAxis, DEGRAD(3.6), tmpXform);
	    PEXMatrixMult(pc_attr.global_transform, tmpXform,
				pc_attr.global_transform);
	    masks[0] = masks[1] = masks[2] = 0;
	    PEXSetPCAttributeMask(masks, PEXPCGlobalTransform);
	    PEXChangePipelineContext(dpy, pc, masks, &pc_attr);
	}

	i = (i > 0) ? --i : (NUM_FRAMES - 1);
	frame = (reverse) ? i : (NUM_FRAMES - 1) - i;

        if (db_data.do_clear) clear_back_buffer( dpy, &db_data);

	PEXBeginRendering(dpy, db_data.back_buffer, rd);
	  PEXExecuteStructure(dpy, rd, PEXOCRender, wavestr[frame]);
	PEXEndRendering(dpy, rd, True);

        display_back_buffer(dpy, &db_data);
	
	XSync(dpy, 0);
    }
    
    /* Termination */
   destroy_back_buffer( dpy, &db_data);
   XCloseDisplay( dpy);
   exit( 0);
}



usage(name)
    char    *name;
{
    fprintf(stderr, "\n%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
	"usage: ", name, " [options]\n\n",
	"    Animation of a 3D wave.\n\n",
	"  Options:\n",
	"    -d displayname      specify an X server connection\n",
	"    -g geometry         specify window geometry in pixels\n",
	"    -s width height     number of polygons used to define surface\n",
	"    -demo count         run for specified number of wave cycles\n",
	"    -r                  reverse the direction of wave propagation\n",
	"    -z                  use z-buffer for hidden surface removal\n"
    );
}
