/*
 *  File Name: polyline_set.c
 *
 *  Description: Polyline set example from Chapter 5
 * 
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"

#define	TITLE 		"PEX Polyline Set"

#define N_VERT 7

/* 
 * Used to draw polyline set
 */
static PEXVertexRGB zigzag_color[N_VERT] = 
{
/* vertex            color             */
 {{0.05, 0.3, 0.0},    {1.0,  0.0, 0.0}},  /* 1st vertex is bright red */
 {{0.2,  0.7, 0.0},    {0.85, 0.0, 0.0}},  /* decrease red for 2nd vertex */
 {{0.35, 0.3, 0.0},    {0.7,  0.0, 0.0}},  /* continue decreasing red */
 {{0.5,  0.7, 0.0},    {0.55, 0.0, 0.0}},  /* 4th vertex is about 50% red */
 {{0.65, 0.3, 0.0},    {0.4,  0.0, 0.0}},  /* decrease red for 5th vertex */
 {{0.8,  0.7, 0.0},    {0.25, 0.0, 0.0}},  /* continue decreasing red */
 {{0.95, 0.3, 0.0},    {0.1,  0.0, 0.0}},  /* last vertex is almost black */
};

/* Draw a polyline set with color interpolation */
int draw_line_set ( dpy, win, rd )
Display *dpy; 
Window win; 
PEXRenderer rd;
{
    PEXListOfVertex pline_set;

    pline_set.count = N_VERT;
    pline_set.vertices.rgb = zigzag_color;

    PEXBeginRendering( dpy, win, rd);
	PEXSetLineWidth( dpy, rd, PEXOCRender, 2.0);
	PEXSetPolylineInterpMethod( dpy, rd, PEXOCRender, 
		PEXPolylineInterpColor);
        PEXPolylineSetWithData( dpy, rd, PEXOCRender, PEXGAColor, 
		PEXColorTypeRGB, 1, &pline_set);
    PEXEndRendering( dpy, rd, True);
}

int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done;   /*indicates when program is done*/

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);
 
   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE, WhitePixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   /* Ask for mouse press and window exposure events */
   XSelectInput( dpy, win, ButtonPressMask | ExposureMask);

   /* If PEX version 5.1 or later then let PEX clear the window to 
    * white before rendering */
   rd_flags = 0;
   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb.red = 1.0;
      rd_attr.background_color.value.rgb.green = 1.0;
      rd_attr.background_color.value.rgb.blue = 1.0;
      rd_flags |= PEXRAClearImage | PEXRABackgroundColor;
   }

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Event loop */
   done = False;
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window if we are talking to a PEX 5.0 server
          *    Redraw the lines
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
            if ( ext_info->minor_version < 1 )
               XClearWindow( dpy, win);
            draw_line_set( dpy, win, rd);
            break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   XCloseDisplay( dpy);
   exit( 0);
}
