/*
 *  File Name: teapot.c
 * 
 *  Description: Teapot program from Chapter 11
 *
 *  Author: Paula Womack
 *
 *  Revision: 1.0
 *
 */

#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "view_utils.h"
#include "util_protos.h"
#include "teapot_data.h"

#define	TITLE1 		"PEX Teapot - No Lighting"
#define	TITLE2 		"PEX Teapot - Ambient Lighting"
#define	TITLE3 		"PEX Teapot - Diffuse Lighting"
#define	TITLE4 		"PEX Teapot - Specular Lighting"
#define	TITLE5 		"PEX Teapot - Depth Cued"

/* Set this to 1 to replace point light with spot light */
#define USE_SPOT 	0

#define MAX_LIGHTS 10 

static PEXColorRGB  red_color = {1.0, 0.0, 0.0};
static PEXColorRGB  background_color = {0.0, 0.0, 0.0};


/*
 * Set up depth cueing 
 */
init_depth_cue( dpy, win, dc_table, dc_index)
Display *dpy; 
Window win; 
PEXLookupTable dc_table;
unsigned int *dc_index;
{
    PEXDepthCueEntry depth_cue_entry;
    
    /* Store dc entry at location one */
    *dc_index = 1;

    /* Have objects tend to background color */
    depth_cue_entry.mode = PEXOn;
    depth_cue_entry.front_plane = 1.0;
    depth_cue_entry.back_plane = 0.0;
    depth_cue_entry.front_scaling = 1.0;
    depth_cue_entry.back_scaling = 0.0;
    depth_cue_entry.color.type = PEXColorTypeRGB;
    depth_cue_entry.color.value.rgb = background_color;
    PEXSetTableEntries( dpy, dc_table, *dc_index, 1/*count*/, PEXLUTDepthCue, 
        &depth_cue_entry);
}


/* 
 * Store some lights
 */
int add_lights( dpy, win, light_table, lights)
Display *dpy; 
Window win; 
PEXLookupTable light_table;
PEXTableIndex *lights;

{
    unsigned short imp_dep_const_name;
    PEXImpDepConstant *imp_dep_const;
    int num_non_ambients;
    int enum_type_name;
    unsigned long *enum_counts;
    PEXEnumTypeDesc *enum_desc;
    int num_enums;
    int do_ambient, do_vector, do_point, do_spot;
    PEXLightEntry light_entry;
    PEXVector light_dir;
    int nlights = 0;

    /* Find out how many non-ambient lights can be enabled at once */
    imp_dep_const_name = PEXIDMaxNonAmbientLights;
    if (!PEXGetImpDepConstants (dpy, win, 1, &imp_dep_const_name, &imp_dep_const) ) {
        fprintf( stderr, "Error calling PEXGetImpDepConstants\n");
        return nlights;
    }
    num_non_ambients = imp_dep_const[0].integer;
    XFree( imp_dep_const);

    /* 
     * Get the supported light types. Start with all light types disabled then loop
     * through the list of supported lights enabling the types that are supported
     */
    enum_type_name = PEXETLightType;
    if (!PEXGetEnumTypeInfo(dpy, win, 1, &enum_type_name, PEXETIndex, 
        &enum_counts, &enum_desc )) {
        fprintf( stderr, "Error calling PEXGetEnumTypeInfo\n");
        return nlights;
    }

    do_ambient = do_vector = do_point = do_spot = False;
    num_enums = enum_counts[0];
    while (--num_enums >= 0) {
        if (enum_desc[num_enums].index == PEXLightAmbient) do_ambient = True;
        if (enum_desc[num_enums].index == PEXLightWCVector) do_vector = True;
        if (enum_desc[num_enums].index == PEXLightWCPoint) do_point = True;
        if (enum_desc[num_enums].index == PEXLightWCSpot) do_spot= True;
    }
    PEXFreeEnumInfo( 1, enum_counts, enum_desc);

    if (USE_SPOT && do_spot)
        do_point = False;
    else
        do_spot = False;

    /* If ambient lights are supported then add a gray ambient light */
    if (do_ambient) {
        light_entry.type = PEXLightAmbient;
        light_entry.color.type = PEXColorTypeRGB;
        light_entry.color.value.rgb.red = 0.4;
        light_entry.color.value.rgb.green = 0.4;
        light_entry.color.value.rgb.blue = 0.4;
        PEXSetTableEntries(dpy, light_table, 1/*entry*/, 1/*count*/,
            PEXLUTLight, &light_entry);
        lights[nlights++] = 1;
    }

    /* 
     * If vector lights are supported and we haven't exceeded the maximum 
     * non-ambients then add a white vector light coming from the front left 
     */ 
    if (do_vector && num_non_ambients > 0) {
        light_entry.type = PEXLightWCVector;
        light_dir.x =  1.0;
        light_dir.y =  0.0;
        light_dir.z = -1.0;
        PEXNormalizeVectors(1/*count*/, &light_dir, &light_entry.direction);
        light_entry.color.type = PEXColorTypeRGB;
        light_entry.color.value.rgb.red = 0.6;
        light_entry.color.value.rgb.green = 0.6;
        light_entry.color.value.rgb.blue = 0.6;
        PEXSetTableEntries(dpy, light_table, 2/*entry*/, 1/*count*/,
            PEXLUTLight, &light_entry);
        lights[nlights++] = 2;
        num_non_ambients--;
    }

    /* 
     * If point lights are supported and we haven't exceeded the maximum 
     * non-ambients then add a white point light overhead
     */ 
    if (do_point && num_non_ambients > 0) {
        light_entry.type = PEXLightWCPoint;
        light_entry.point.x = (PointMaxX - PointMinX)/2.0 + PointMinX;
        light_entry.point.y = PointMaxY + (PointMaxY - PointMinY)/10.0;
        light_entry.point.z = (PointMaxZ - PointMinZ)/2.0 + PointMinZ;
        light_entry.color.type = PEXColorTypeRGB;
        light_entry.color.value.rgb.red = 0.6;
        light_entry.color.value.rgb.green = 0.6;
        light_entry.color.value.rgb.blue = 0.6;
        light_entry.attenuation1 = 1.0;
        light_entry.attenuation2 = 5.0;
        PEXSetTableEntries(dpy, light_table, 3/*entry*/, 1/*count*/,
            PEXLUTLight, &light_entry);
        lights[nlights++] = 3;
        num_non_ambients--;
    }

    /* 
     * If spot lights are supported and we haven't exceeded the maximum 
     * non-ambients then add a white point light overhead (the code will
     * only do a spot light or a point light so we can just use entry 3)
     */ 
    if (do_spot && num_non_ambients > 0) {
        light_entry.type = PEXLightWCSpot;
        light_entry.point.x = (PointMaxX - PointMinX)/2.0 + PointMinX;
        light_entry.point.y = PointMaxY + (PointMaxY - PointMinY)/10.0;
        light_entry.point.z = (PointMaxZ - PointMinZ)/2.0 + PointMinZ;
        light_entry.color.type = PEXColorTypeRGB;
        light_entry.color.value.rgb.red = 0.6;
        light_entry.color.value.rgb.green = 0.6;
        light_entry.color.value.rgb.blue = 0.6;
        light_entry.direction.x = 0.;
        light_entry.direction.y = -1.;
        light_entry.direction.z = 0.;
        light_entry.attenuation1 = 1.0;
        light_entry.attenuation2 = 5.0;
        light_entry.spread_angle = M_PI/4.0;
        light_entry.concentration = 2.0;
        PEXSetTableEntries(dpy, light_table, 3/*entry*/, 1/*count*/,
            PEXLUTLight, &light_entry);
        lights[nlights++] = 3;
        num_non_ambients--;
    }

    return nlights;
}

PEXStructure store_teapot( Display *dpy)
{
    extern PEXVector fnormals[];
    extern PEXVector vnormals[];
    extern PEXCoord  vpoints[];

    int i, j, k;
    PEXArrayOfVertex  vertices;
    PEXFacetData      facet_data;
    PEXVertexNormal   vertex[3];
    PEXStructure sid;

    sid = PEXCreateStructure( dpy);

    vertices.normal = vertex;
    for ( i = 0, j = 0; i < NumFacets; i++, j+=3 ) 
    {
	facet_data.normal = fnormals[i];
	for (k = 0; k < 3; k++)
	{
	    vertex[k].normal = vnormals[j+k];
	    vertex[k].point  = vpoints[j+k];
	}
	PEXFillAreaWithData( dpy, sid, PEXOCStore, PEXShapeConvex, False, 
		PEXGANormal, PEXGANormal, 0/*no colors*/, 
		&facet_data, 3, vertices); 
    }


    return sid;
}

int draw_dc_teapot( dpy, win, rd, sid, dc_index )
Display *dpy; 
Window win; 
PEXRenderer rd; 
PEXStructure sid;
unsigned int dc_index;
{
    PEXMatrix xrot_matrix, yrot_matrix, rot_matrix;

     /*
      * Get rotation matrix (object data is centered about
      * (0,0) so translation is not necessary)
      */
    PEXRotate (PEXXAxis, M_PI/6.0, xrot_matrix);
    PEXRotate (PEXYAxis, M_PI/6.0, yrot_matrix);
    PEXMatrixMult (xrot_matrix, yrot_matrix, rot_matrix);
 
    PEXBeginRendering( dpy, win, rd);
        PEXSetSurfaceColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
            (PEXColor *)&red_color);
        PEXSetInteriorStyle( dpy, rd, PEXOCRender, PEXInteriorStyleHollow);
        PEXSetDepthCueIndex ( dpy, rd, PEXOCRender, dc_index);
        PEXSetLocalTransform( dpy, rd, PEXOCRender, PEXReplace, rot_matrix);
        PEXExecuteStructure( dpy, rd, PEXOCRender, sid);
    PEXEndRendering( dpy, rd, True);
}

int draw_teapot( dpy, win, rd, sid, reflection, num_lights, lights )
Display *dpy; 
Window win; 
PEXRenderer rd; 
PEXStructure sid;
int reflection;
int num_lights;
PEXTableIndex *lights;
{
    PEXReflectionAttributes reflection_attr;
    PEXMatrix xrot_matrix;

     /*
      * Get rotation matrix (object data is centered about
      * (0,0) so translation is not necessary
      */
    PEXRotate (PEXXAxis, M_PI/18.0, xrot_matrix);
 
    PEXBeginRendering( dpy, win, rd);
        PEXSetSurfaceColor( dpy, rd, PEXOCRender, PEXColorTypeRGB, 
            (PEXColor *)&red_color);
        PEXSetLightSourceState (dpy, rd, PEXOCRender, num_lights, lights, 0, NULL);
        PEXSetInteriorStyle( dpy, rd, PEXOCRender, PEXInteriorStyleSolid);
        PEXSetSurfaceInterpMethod( dpy, rd, PEXOCRender, PEXSurfaceInterpColor);
        PEXSetReflectionModel( dpy, rd, PEXOCRender, reflection);

        if (reflection == PEXReflectionAmbient)
            reflection_attr.ambient = 1.0;
        else
            reflection_attr.ambient = 0.8;
        if (reflection == PEXReflectionDiffuse)
            reflection_attr.diffuse = 1.0;
        else
            reflection_attr.diffuse = 0.8;
        reflection_attr.specular = 1.0;
        reflection_attr.specular_conc = 10.0;
        reflection_attr.transmission = 0.0;
        reflection_attr.specular_color.type = PEXColorTypeRGB;
        reflection_attr.specular_color.value.rgb.red = 1.0;
        reflection_attr.specular_color.value.rgb.green = 1.0;
        reflection_attr.specular_color.value.rgb.blue = 1.0;
        PEXSetReflectionAttributes( dpy, rd, PEXOCRender, &reflection_attr);

        PEXSetFacetCullingMode( dpy, rd, PEXOCRender, PEXBackFaces);
        PEXSetLocalTransform( dpy, rd, PEXOCRender, PEXReplace, xrot_matrix);
        PEXExecuteStructure( dpy, rd, PEXOCRender, sid);

    PEXEndRendering( dpy, rd, True);
}


int main(argc, argv)
int argc; 
char *argv[];
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XSizeHints			win_hint; /* window manager hints */
   XVisualInfo 			vis_info; /* color information */
   XStandardColormap 		cmap_info;
   PEXColorApproxEntry         	approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done;   /*indicates when program is done*/
   char 			key_pressed;
   int				demo;
   PEXStructure                 sid;
   unsigned int		        dc_index;
   int 				num_lights;
   PEXTableIndex 	        lights[MAX_LIGHTS];
   int 				subset;

   /* Open a connection to the PEX server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);
    
   /* Make sure the structure subset is supported */
   subset = ext_info->subset_info & 0xff;
   if ( (subset != 0) && !(subset & PEXStructureMode) ) {
      fprintf( stderr, "Structures are not supported\n");
      exit( 1);
   }

   /* Find best visual and colormap. If we fail, just exit */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);

   /* Create a window */
   win = create_window( dpy, TITLE1, BlackPixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   /* Ask for key press, mouse press and window exposure events */
   XSelectInput( dpy, win, KeyPressMask | ButtonPressMask | ExposureMask);

   /* Create PEX view such that object is in center */
   rd_attr.view_table = PEXCreateLookupTable( dpy, win, PEXLUTView);
   rd_flags = PEXRAViewTable;
   init_object_view( dpy, PointMinX, PointMaxX, PointMinY, PointMaxY, PointMinZ,
      PointMaxZ, 1.0/*scale*/, ORTHOGRAPHIC_PROJECTION, FRONT_VIEW, 
      0/*default entry*/, rd_attr.view_table);

   /* Set up depth cueing at table entry 1 */
   rd_attr.depth_cue_table = PEXCreateLookupTable( dpy, win, PEXLUTDepthCue);
   rd_flags |= PEXRADepthCueTable;
   init_depth_cue( dpy, win, rd_attr.depth_cue_table, &dc_index);

   /* Enable Z-buffering */
   rd_attr.hlhsr_mode = PEXHLHSRZBuffer;
   rd_flags |= PEXRAHLHSRMode;

   /* Add some lights */
   rd_attr.light_table = PEXCreateLookupTable( dpy, win, PEXLUTLight);
   rd_flags |= PEXRALightTable;
   num_lights = add_lights( dpy, win, rd_attr.light_table, lights);

   /* If PEX version 5.1 or later then let PEX clear the window to 
    * white before rendering */
   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb = background_color;
      rd_flags |= PEXRAClearImage | PEXRABackgroundColor;
   }

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Store the teapot in a structure */
   sid = store_teapot( dpy);

   /* Event loop */
   win_hint.flags = 0;
   done = False;
   demo = 1; 
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window if we are talking to a PEX 5.0 server
          *    Redraw the teapot
          */
	 case Expose: 
            while ( XCheckTypedWindowEvent( dpy, win, Expose, &event) );
            if ( ext_info->minor_version < 1 )
               XClearWindow( dpy, win);
            draw_teapot( dpy, win, rd, sid, PEXReflectionNone, num_lights, lights);
            break;

	/*
         * Show the next reflection model whenever 'n' or 'N' is typed
         */
         case KeyPress:
	    XLookupString((XKeyEvent *)&event, &key_pressed, 1, NULL, NULL);
	    if (key_pressed == 'n' || key_pressed == 'N') {
                demo++; 
                if (demo > 5) demo = 1;

                if ( ext_info->minor_version < 1 )
                    XClearWindow( dpy, win);

		if (demo == 1) {
   		    XSetStandardProperties( dpy, win, TITLE1, TITLE1, 
				None, argv, argc, &win_hint);
		    draw_teapot( dpy, win, rd, sid, PEXReflectionNone, num_lights, 
		        lights);
	  	} else if (demo == 2) {
   		    XSetStandardProperties( dpy, win, TITLE2, TITLE2, 
				None, argv, argc, &win_hint);
		    draw_teapot( dpy, win, rd, sid, PEXReflectionAmbient, num_lights, 
			lights);
                } else  if (demo == 3) {
   		    XSetStandardProperties( dpy, win, TITLE3, TITLE3, 
				None, argv, argc, &win_hint);
		    draw_teapot( dpy, win, rd, sid, PEXReflectionDiffuse, num_lights, 
			lights);
                } else if (demo == 4) {
   		    XSetStandardProperties( dpy, win, TITLE4, TITLE4, 
				None, argv, argc, &win_hint);
		    draw_teapot( dpy, win, rd, sid, PEXReflectionSpecular, num_lights, 
			lights);
                } else /* demo == 5 */ {
   		    XSetStandardProperties( dpy, win, TITLE5, TITLE5, 
				None, argv, argc, &win_hint);
		    draw_dc_teapot( dpy, win, rd, sid, dc_index);
                }
	    }

	    break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   XCloseDisplay( dpy);
   exit( 0);
}
