/*
 *  File Name: indexed_color_lines.c
 * 
 *  Author: Paula Womack
 *
 * Description:
 *    Contains code from Chapter 4 which sets up indexed colors.
 *    Contains polyline set code in Chapter 5.
 *    Contains quadmesh code in Chapter 6.
 *
 * Revision: 1.0
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/PEX5/PEXlib.h>
#include "util_protos.h"

#define	TITLE 		"PEX Indexed Colors"

#define N_LINE 5
static PEXCoord diagonal1[N_LINE] =
{
 {0.1,   0.1,  0.0},
 {0.2,   0.1,  0.0},
 {0.9,   0.9,  0.0},
 {0.8,   0.9,  0.0},
 {0.1,   0.1,  0.0},
};
static PEXCoord diagonal2[N_LINE] =
{
 {0.1,   0.9,  0.0},
 {0.2,   0.9,  0.0},
 {0.9,   0.1,  0.0},
 {0.8,   0.1,  0.0},
 {0.1,   0.9,  0.0},
};

/*
 * Code from Chapter 4 to set up PEX color table with 0-6 colors
 */
#define NUM_COLOR 6
PEXColorRGB colors[NUM_COLOR] = 
{
    {1.0, 0.0, 0.0},   /* red */
    {0.0, 1.0, 0.0},   /* green */
    {0.0, 0.0, 1.0},   /* blue */
    {1.0, 1.0, 0.0},   /* yellow */
    {1.0, 0.0, 1.0},   /* magenta */
    {0.0, 1.0, 1.0},   /* cyan */
};

int setup_indexed_colors(Display *dpy, Window win, PEXRenderer rd, int *start, 
		      int *num)
{
    PEXTableInfo clrTableInfo;
    PEXRendererAttributes rd_attr;
    PEXColorEntry colorEntries[NUM_COLOR];
    int numEntries, i;

    for (i = 0; i < NUM_COLOR; i++) {
        colorEntries[i].type = PEXColorTypeRGB;
        colorEntries[i].value.rgb = colors[i];
    }
    
    PEXGetTableInfo( dpy, win, PEXLUTColor, &clrTableInfo);
    rd_attr.color_table = PEXCreateLookupTable( dpy, win, PEXLUTColor);
    numEntries = clrTableInfo.definable_entries - clrTableInfo.predefined_count;
    if (numEntries > NUM_COLOR) numEntries = NUM_COLOR;
    PEXSetTableEntries( dpy, rd_attr.color_table, clrTableInfo.predefined_max + 1,
        numEntries, PEXLUTColor, colorEntries);
    PEXChangeRenderer( dpy, rd, PEXRAColorTable, &rd_attr);

    *start = clrTableInfo.predefined_max + 1;
    *num = numEntries;
}


/*
 * Code from Chapter 5 to use polyline sets
 */
int draw_lines( Display *dpy, Window win, PEXRenderer rd, int index )
{
    PEXListOfVertex pline_set[2];

    pline_set[0].count = pline_set[1].count = N_LINE;
    pline_set[0].vertices.no_data = diagonal1;
    pline_set[1].vertices.no_data = diagonal2;

    PEXBeginRendering( dpy, win, rd);
	PEXSetLineWidth( dpy, rd, PEXOCRender, 2.0);
        PEXSetLineColorIndex( dpy, rd, PEXOCRender, index);
        PEXPolylineSetWithData( dpy, rd, PEXOCRender, PEXGANone, 0, 2, 
            pline_set);
    PEXEndRendering( dpy, rd, True);
}


/*
 * Code from Chapter 6 to use quadmesh
 */
int draw_quadmesh( Display *dpy, Window win, PEXRenderer rd)
{
#define NUM_ROW 3
#define NUM_COL 4
    PEXArrayOfVertex vertices;
    PEXArrayOfFacetData facet_data;
    PEXCoord quad_verts[NUM_ROW][NUM_COL];
    PEXColorRGB facet_colors[NUM_ROW-1][NUM_COL-1];
    int	col_count, row_count, i, j;
    PEXCoord start_vert, next_vert;

    row_count = NUM_ROW; col_count = NUM_COL;
    start_vert.x = start_vert.y = 0.2; start_vert.z = 0.0;
    next_vert = start_vert;

    for (i = 0; i < row_count; i++) {
        for (j = 0; j < col_count; j++) {
            quad_verts[i][j] = next_vert;
            next_vert.x += 0.2;
        }
        next_vert.x  = start_vert.x;
        next_vert.y += 0.3;
    }
    vertices.no_data = (PEXCoord *) quad_verts;

    for (i = 0; i < row_count - 1; i++)
        for (j = 0; j < col_count - 1; j++)
            facet_colors[i][j] = colors[i*(col_count-1) + j];
    facet_data.rgb = (PEXColorRGB *)facet_colors;

    PEXBeginRendering( dpy, win, rd);
        PEXSetInteriorStyle( dpy, rd, PEXOCRender, PEXInteriorStyleSolid);
        PEXQuadrilateralMesh( dpy, rd, PEXOCRender,  PEXShapeConvex, PEXGAColor, 
            PEXGANone, PEXColorTypeRGB, facet_data, col_count, row_count, vertices);
    PEXEndRendering( dpy, rd, True);
}


int main(int argc, char *argv[])
{
   Display			*dpy;    /* connection information */
   Window			win;     /* window to draw in */
   unsigned int			width, height;
   XEvent	    		event;   /* event record returned by X */
   XVisualInfo                  vis_info; /* color information */
   XStandardColormap            cmap_info;
   PEXColorApproxEntry          approx_entry;
   PEXRenderer			rd;      /* PEX renderer resource */
   PEXRendererAttributes	rd_attr;
   unsigned long		rd_flags;
   PEXExtensionInfo 		*ext_info;/* info about PEX extension */
   int				done;   /*indicates when program is done*/
   char                         key_pressed;
   int				next_index, start_index, num_indices;

   /* Open a connection to the X server */
   dpy = open_pex_connection( &ext_info);
   if ( dpy == NULL ) exit( 1);

   /* Initialize the PEX extension and get the PEX extension information */
   /* We will need to know the version of PEX later on.                  */
   if ( !setup_colors( dpy, &vis_info, &cmap_info, &approx_entry) )
      exit( 1);
       
   /* Compute the window size so it is about 50% of the screen size */
   win = create_window( dpy, TITLE, WhitePixel( dpy, DefaultScreen( dpy)),
           &vis_info, &cmap_info, &width, &height);

   /* Ask for mouse press and window exposure events */
   XSelectInput( dpy, win,  KeyPressMask | ButtonPressMask | ExposureMask);

   /* If PEX version 5.1 or later then let PEX clear the window to 
    * white before rendering */
   rd_flags = 0;
   if ( ext_info->minor_version > 0 ) {
      rd_attr.clear_image = True;
      rd_attr.background_color.type = PEXColorTypeRGB;
      rd_attr.background_color.value.rgb.red = 1.0;
      rd_attr.background_color.value.rgb.green = 1.0;
      rd_attr.background_color.value.rgb.blue = 1.0;
      rd_flags |= PEXRAClearImage | PEXRABackgroundColor;
   }

   /* Create a PEX renderer with the attributes specified above */
   rd = create_renderer( dpy, win, &approx_entry, rd_flags, &rd_attr);

   /* Map the window */
   XMapWindow( dpy, win);

   /* Store indexed colors */
   setup_indexed_colors(dpy, win, rd, &start_index, &num_indices);
   if (num_indices == 0) {
      fprintf( stderr, "Can't define any indexed colors");
      exit( 1);
   }

   /* Event loop */
   done = False;
   next_index = start_index;
   while ( !done ) {
      XNextEvent( dpy, &event);
      switch ( event.type ) {
         /* 
          * Whenever we get an expose event we must:
          *    Get rid of all other expose events on the queue
          *    Clear the window if we are talking to a PEX 5.0 server
          *    Redraw the lines
          */
	 case Expose: 
            while ( XCheckTypedEvent( dpy, Expose, &event) );
            if ( ext_info->minor_version < 1 )
               XClearWindow( dpy, win);
            draw_lines( dpy, win, rd, next_index);
            break;

        /*
         * Use the next indexed color whenever 'n' or 'N' is typed
         */
         case KeyPress:
            XLookupString( (XKeyEvent *)&event, &key_pressed, 1, NULL, NULL);
            if (key_pressed == 'n' || key_pressed == 'N') {
                next_index++;
                if (next_index >= start_index + num_indices) 
	            next_index = start_index;

                if ( ext_info->minor_version < 1 )
                    XClearWindow( dpy, win);
                draw_lines( dpy, win, rd, next_index);
	    } else if (key_pressed == 'q' || key_pressed == 'Q') {
                if ( ext_info->minor_version < 1 )
                    XClearWindow( dpy, win);
                draw_quadmesh( dpy, win, rd);
	    }
	    break;

         /* Quit when mouse button is pressed */
	 case ButtonPress:
	    done = True;
	    break;
      }
   }

   /* Termination */
   PEXFreeRenderer( dpy, rd);
   XDestroyWindow( dpy, win);
   XCloseDisplay( dpy);
   exit( 0);
}
